#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import date
from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot._compat import TypedDict
from datarobot.models.api_object import APIObject
from datarobot.utils import from_api
from datarobot.utils.pagination import unpaginate


class LLMReferenceDict(TypedDict):
    """Dict representation of LlmReference."""

    name: str
    url: Optional[str]


class LLMGatewayCatalogEntryDict(TypedDict):
    """Dict representation of LlmGatewayCatalogEntry."""

    model: str
    llm_id: str
    name: str
    description: str
    provider: str
    creator: str
    context_size: int
    max_completion_tokens: int
    capabilities: Optional[List[str]]
    supported_languages: List[str]
    input_types: List[str]
    output_types: List[str]
    parameters: Dict[str, Any]
    documentation_link: str
    reference_links: List[LLMReferenceDict]
    date_added: date
    license: str
    is_preview: bool
    is_metered: bool
    retirement_date: Optional[date]
    suggested_replacement: Optional[str]
    is_deprecated: bool
    is_active: bool
    available_regions: List[str]


llm_reference_trafaret = t.Dict(
    {
        t.Key("name"): t.String,
        t.Key("url", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")

llm_gateway_catalog_entry_trafaret = t.Dict(
    {
        t.Key("model"): t.String,
        t.Key("llm_id"): t.String,
        t.Key("name"): t.String,
        t.Key("description"): t.String,
        t.Key("provider"): t.String,
        t.Key("creator"): t.String,
        t.Key("context_size"): t.Int,
        t.Key("max_completion_tokens"): t.Int,
        t.Key("capabilities", optional=True): t.Or(t.List(t.String), t.Null),
        t.Key("supported_languages"): t.List(t.String),
        t.Key("input_types"): t.List(t.String),
        t.Key("output_types"): t.List(t.String),
        t.Key("parameters"): t.Dict({}).allow_extra("*"),
        t.Key("documentation_link"): t.String,
        t.Key("reference_links"): t.List(llm_reference_trafaret),
        t.Key("date_added"): t.String,
        t.Key("license"): t.String,
        t.Key("is_preview"): t.Bool,
        t.Key("is_metered"): t.Bool,
        t.Key("retirement_date", optional=True): t.Or(t.String, t.Null),
        t.Key("suggested_replacement", optional=True): t.Or(t.String, t.Null),
        t.Key("is_deprecated"): t.Bool,
        t.Key("is_active"): t.Bool,
        t.Key("available_regions"): t.List(t.String),
    }
).ignore_extra("*")


class LLMReference(APIObject):
    """
    Reference link for an LLM.

    Attributes
    ----------
    name : str
        Description of the reference document.
    url : str or None
        URL of the reference document.
    """

    _converter = llm_reference_trafaret

    def __init__(self, name: str, url: Optional[str] = None):
        self.name = name
        self.url = url

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name})"

    def to_dict(self) -> LLMReferenceDict:
        return {"name": self.name, "url": self.url}


class LLMGatewayCatalogEntry(APIObject):
    """
    Metadata for an LLM Gateway catalog entry.

    Attributes
    ----------
    model : str
        The identifier of the model to use for chat completion.
    llm_id : str
        The internal identifier of the LLM.
    name : str
        The name of the LLM.
    description : str
        The description of the LLM.
    provider : str
        The name of the LLM provider.
    creator : str
        The creator of the LLM.
    context_size : int
        The context size of the LLM.
    max_completion_tokens : int
        The maximum number of tokens the LLM can generate.
    capabilities : list of str or None
        The capabilities of the LLM.
    supported_languages : list of str
        The languages supported by the LLM.
    input_types : list of str
        The input types supported by the LLM.
    output_types : list of str
        The supported output types by the LLM.
    parameters : dict
        The supported LLM completion parameters.
    documentation_link : str
        The link to the LLM documentation.
    reference_links : list of LlmReference
        The reference links.
    date_added : date
        The date the LLM was added to the catalog.
    license : str
        The license of the LLM.
    is_preview : bool
        Whether the LLM is part of fast-track LLMs.
    is_metered : bool
        Whether the LLM is metered for consumption-based pricing.
    retirement_date : date or None
        The date the LLM was/will be retired.
    suggested_replacement : str or None
        The suggested replacement for the LLM.
    is_deprecated : bool
        Whether the LLM is deprecated.
    is_active : bool
        Whether the LLM is active.
    available_regions : list of str
        The regions where the LLM is available.
    """

    _converter = llm_gateway_catalog_entry_trafaret

    def __init__(
        self,
        model: str,
        llm_id: str,
        name: str,
        description: str,
        provider: str,
        creator: str,
        context_size: int,
        max_completion_tokens: int,
        supported_languages: List[str],
        input_types: List[str],
        output_types: List[str],
        parameters: Dict[str, Any],
        documentation_link: str,
        reference_links: List[Dict[str, Any]],
        date_added: str,
        license: str,
        is_preview: bool,
        is_metered: bool,
        is_deprecated: bool,
        is_active: bool,
        available_regions: List[str],
        capabilities: Optional[List[str]] = None,
        retirement_date: Optional[str] = None,
        suggested_replacement: Optional[str] = None,
    ):
        self.model = model
        self.llm_id = llm_id
        self.name = name
        self.description = description
        self.provider = provider
        self.creator = creator
        self.context_size = context_size
        self.max_completion_tokens = max_completion_tokens
        self.capabilities = capabilities
        self.supported_languages = supported_languages
        self.input_types = input_types
        self.output_types = output_types
        self.parameters = parameters
        self.documentation_link = documentation_link
        self.reference_links = [LLMReference.from_server_data(ref) for ref in reference_links]
        self.date_added = (
            date.fromisoformat(date_added) if isinstance(date_added, str) else date_added
        )
        self.license = license
        self.is_preview = is_preview
        self.is_metered = is_metered
        self.retirement_date = date.fromisoformat(retirement_date) if retirement_date else None
        self.suggested_replacement = suggested_replacement
        self.is_deprecated = is_deprecated
        self.is_active = is_active
        self.available_regions = available_regions

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(model={self.model}, name={self.name})"

    def to_dict(self) -> LLMGatewayCatalogEntryDict:
        return {
            "model": self.model,
            "llm_id": self.llm_id,
            "name": self.name,
            "description": self.description,
            "provider": self.provider,
            "creator": self.creator,
            "context_size": self.context_size,
            "max_completion_tokens": self.max_completion_tokens,
            "capabilities": self.capabilities,
            "supported_languages": self.supported_languages,
            "input_types": self.input_types,
            "output_types": self.output_types,
            "parameters": self.parameters,
            "documentation_link": self.documentation_link,
            "reference_links": [ref.to_dict() for ref in self.reference_links],
            "date_added": self.date_added,
            "license": self.license,
            "is_preview": self.is_preview,
            "is_metered": self.is_metered,
            "retirement_date": self.retirement_date,
            "suggested_replacement": self.suggested_replacement,
            "is_deprecated": self.is_deprecated,
            "is_active": self.is_active,
            "available_regions": self.available_regions,
        }


class LLMGatewayCatalog(APIObject):
    """
    LLM Gateway catalog management for listing and validating available LLMs.

    This class provides convenient methods for interacting with the LLM Gateway catalog,
    including filtering for active, non-deprecated models and validating model availability.
    """

    _path = "genai/llmgw/catalog/"

    @classmethod
    def list(
        cls,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        llm_id: Optional[str] = None,
        only_active: bool = True,
        only_non_deprecated: bool = True,
    ) -> List[LLMGatewayCatalogEntry]:
        """
        List LLMs available in the LLM Gateway catalog.

        Parameters
        ----------
        offset : int, optional
            Skip the specified number of values.
        limit : int, optional
            Retrieve only the specified number of values.
        llm_id : str, optional
            Get the catalog entries for a specific LLM ID.
        only_active : bool, default True
            If True, only return active LLMs.
        only_non_deprecated : bool, default True
            If True, only return non-deprecated LLMs.

        Returns
        -------
        list of LlmGatewayCatalogEntry
            A list of LLM Gateway catalog entries.

        Examples
        --------
        .. code-block:: python

            import datarobot as dr

            # List all active, non-deprecated LLMs
            llms = dr.LlmGatewayCatalog.list()

            # List all LLMs (including deprecated and inactive)
            all_llms = dr.LlmGatewayCatalog.list(
                only_active=False,
                only_non_deprecated=False
            )

            # Get specific LLM
            gpt4_models = dr.LlmGatewayCatalog.list(llm_id="azure-openai-gpt-4")

            # Get as dictionaries instead
            llms_dict = dr.LlmGatewayCatalog.list_as_dict()
        """
        params: Dict[str, Union[int, str]] = {}
        if offset is not None:
            params["offset"] = offset
        if limit is not None:
            params["limit"] = limit
        if llm_id is not None:
            params["llmId"] = llm_id

        url = f"{cls._client.domain}/api/v2/{cls._path}"
        r_data = unpaginate(url, params, cls._client)

        # Create catalog entry objects
        catalog_entries = []
        for item in r_data:
            # Convert camelCase to snake_case for the constructor
            converted_data = from_api(item)
            catalog_entries.append(LLMGatewayCatalogEntry.from_server_data(converted_data))

        # Apply filtering
        if only_active:
            catalog_entries = [entry for entry in catalog_entries if entry.is_active]
        if only_non_deprecated:
            catalog_entries = [entry for entry in catalog_entries if not entry.is_deprecated]

        return catalog_entries

    @classmethod
    def list_as_dict(
        cls,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        llm_id: Optional[str] = None,
        only_active: bool = True,
        only_non_deprecated: bool = True,
    ) -> List[LLMGatewayCatalogEntryDict]:
        """
        List LLMs available in the LLM Gateway catalog as dictionaries.

        This is a type-safe version that returns dictionaries instead of objects.

        Parameters
        ----------
        offset : int, optional
            Skip the specified number of values.
        limit : int, optional
            Retrieve only the specified number of values.
        llm_id : str, optional
            Get the catalog entries for a specific LLM ID.
        only_active : bool, default True
            If True, only return active LLMs.
        only_non_deprecated : bool, default True
            If True, only return non-deprecated LLMs.

        Returns
        -------
        list of LlmGatewayCatalogEntryDict
            A list of LLM Gateway catalog entries as dictionaries.

        Examples
        --------
        .. code-block:: python

            import datarobot as dr

            # Get as dictionaries for raw data access
            llms_dict = dr.LlmGatewayCatalog.list_as_dict()
            for llm in llms_dict:
                print(f"Model: {llm['model']}, Active: {llm['is_active']}")
        """
        entries = cls.list(
            offset=offset,
            limit=limit,
            llm_id=llm_id,
            only_active=only_active,
            only_non_deprecated=only_non_deprecated,
        )
        return [entry.to_dict() for entry in entries]

    @classmethod
    def get_available_models(
        cls,
        include_preview: bool = False,
        only_active: bool = True,
        only_non_deprecated: bool = True,
    ) -> List[str]:
        """
        Get a list of available model identifiers.

        Parameters
        ----------
        include_preview : bool, default False
            If True, include preview models.
        only_active : bool, default True
            If True, only return active models.
        only_non_deprecated : bool, default True
            If True, only return non-deprecated models.

        Returns
        -------
        list of str
            A list of available model identifiers.

        Examples
        --------
        .. code-block:: python

            import datarobot as dr

            # Get production-ready models
            models = dr.LlmGatewayCatalog.get_available_models()

            # Get all models including preview
            all_models = dr.LlmGatewayCatalog.get_available_models(
                include_preview=True,
                only_active=False,
                only_non_deprecated=False
            )
        """
        catalog_entries = cls.list(
            only_active=only_active,
            only_non_deprecated=only_non_deprecated,
        )

        if not include_preview:
            catalog_entries = [entry for entry in catalog_entries if not entry.is_preview]

        return [entry.model for entry in catalog_entries]

    @classmethod
    def verify_model_availability(cls, model_id: str) -> LLMGatewayCatalogEntry:
        """
        Validate that a model is available and active in the catalog.

        Parameters
        ----------
        model_id : str
            The model identifier to validate. Can be either 'model' or 'llmId'.

        Returns
        -------
        LlmGatewayCatalogEntry
            The catalog entry for the validated model.

        Raises
        ------
        ValueError
            If the model is not found, not active, or is deprecated.

        Examples
        --------
        .. code-block:: python

            import datarobot as dr

            # Validate a model is available
            try:
                model_entry = dr.LlmGatewayCatalog.verify_model_availability("azure-openai-gpt-4")
                print(f"Model {model_entry.name} is available")
            except ValueError as e:
                print(f"Model validation failed: {e}")
        """
        # Get all models (including inactive and deprecated for validation)
        all_models = cls.list(only_active=False, only_non_deprecated=False)

        # Find matching models by either model or llm_id
        matched_models = [model for model in all_models if model_id in (model.model, model.llm_id)]

        if not matched_models:
            # Get list of available models for error message
            available_models = cls.get_available_models(include_preview=True)
            model_list = "\n.   - ".join(available_models)
            raise ValueError(
                f"Model '{model_id}' not found in catalog. Available models: {model_list}"
            )

        if len(matched_models) > 1:
            raise ValueError(f"Multiple models found for '{model_id}' in catalog. {matched_models}")

        model_entry = matched_models[0]

        if not model_entry.is_active or model_entry.is_deprecated:
            # Get list of available models for error message
            available_models = cls.get_available_models(include_preview=True)
            model_list = "\n.   - ".join(available_models)
            raise ValueError(
                f"Model '{model_id}' is not active or is deprecated. Available models: {model_list}"
            )

        return model_entry

    @classmethod
    def get_model_by_id(cls, model_id: str) -> Optional[LLMGatewayCatalogEntry]:
        """
        Get a specific model by its identifier.

        Parameters
        ----------
        model_id : str
            The model identifier. Can be either 'model' or 'llmId'.

        Returns
        -------
        LlmGatewayCatalogEntry or None
            The catalog entry for the model, or None if not found.

        Examples
        --------
        .. code-block:: python

            import datarobot as dr

            model_entry = dr.LlmGatewayCatalog.get_model_by_id("azure-openai-gpt-4")
            if model_entry:
                print(f"Found model: {model_entry.name}")
            else:
                print("Model not found")
        """
        try:
            model_entry = cls.verify_model_availability(model_id)
            return model_entry
        except ValueError:
            return None



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Trafaret validators that doesn't part of the class.
Separated to own file to improve readability and reusability.
"""
from __future__ import annotations

import trafaret as t

from datarobot._compat import Int, String

# Representation of a single feature impact
# featureName - feature we are describing
# impactUnnormalized - drop in the metric value comparing to base
# impactNormalized - normalized value so that highest drop = 1.0
# redundantWith - one of 2 options
# - None if redundant detection wasn't run or
# - feature name that makes this one redundant
single_feature_impact_trafaret = t.Dict(
    {
        t.Key("impactUnnormalized"): t.Float,
        t.Key("impactNormalized"): t.Float,
        t.Key("redundantWith"): t.Or(t.Null, String),
        t.Key("featureName"): String,
    }
).ignore_extra("*")

# Feature impact data for regression, binary classification and aggregated score for multiclass
# ranRedundancyDetection - shows if redundancy detection was run for the model
# featureImpacts - list of impacts for each feature
feature_impact_trafaret = t.Dict(
    {
        t.Key("count"): Int,
        t.Key("ranRedundancyDetection"): t.Bool,
        t.Key("rowCount"): t.Or(Int, t.Null),
        t.Key("shapBased"): t.Bool,
        # to make newer client work with older DataRobot responses
        t.Key("backtest", optional=True, default=None): t.Or(Int(gte=0), t.Null, t.Atom("holdout")),
        t.Key("featureImpacts"): t.List(single_feature_impact_trafaret),
        t.Key("dataSliceId", optional=True, default=None): t.Or(String, t.Null),
    }
).ignore_extra("*")


custom_model_feature_impact_trafaret = t.Dict(
    {
        t.Key("count"): Int,
        t.Key("ranRedundancyDetection"): t.Bool,
        t.Key("rowCount"): t.Or(Int, t.Null),
        t.Key("shapBased"): t.Bool,
        t.Key("featureImpacts"): t.List(single_feature_impact_trafaret),
    }
).ignore_extra("*")


# Feature impact data for each class in multiclass model
# classFeatureImpacts - list of records for each class, with 2 keys
# - class - target class name
# - featureImpacts - list of impacts of each feature for the class
multiclass_feature_impact_trafaret = t.Dict(
    {
        t.Key("ranRedundancyDetection"): t.Bool,
        t.Key("classFeatureImpacts"): t.List(
            t.Dict(
                {
                    t.Key("featureImpacts"): t.List(single_feature_impact_trafaret),
                    t.Key("class"): String,
                }
            ).ignore_extra("*")
        ),
    }
).ignore_extra("*")



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.notebooks.enums import KernelSpec, KernelState, RuntimeLanguage

notebook_kernel_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.Enum(*list(RuntimeLanguage)),
        t.Key("language"): t.String,
        t.Key("running"): t.Bool,
        t.Key("execution_state"): t.Enum(*list(KernelState)),
    }
).ignore_extra("*")


class NotebookKernel(APIObject):
    """
    A kernel associated with a codespace notebook.

    Attributes
    ----------

    id : str
        The kernel ID.
    name : str
        The kernel name.
    language : RuntimeLanguage
        The kernel language. Supports Python and R.
    running : bool
        Whether the kernel is running.
    execution_state : KernelState
        The kernel execution state.
    """

    _path = "notebookSessions/"

    _converter = notebook_kernel_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        language: RuntimeLanguage,
        running: bool,
        execution_state: KernelState,
    ):
        self.id = id
        self.name = name
        self.language = language
        self.running = running
        self.execution_state = execution_state

    @classmethod
    def create(cls, notebook_id: str, kernel_spec: KernelSpec) -> NotebookKernel:
        r_data = cls._client.post(f"{cls._path}{notebook_id}/kernels/", data={"spec": kernel_spec})
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, notebook_id: str, kernel_id: str) -> NotebookKernel:
        r_data = cls._client.get(f"{cls._path}{notebook_id}/kernels/{kernel_id}/")
        return cls.from_server_data(r_data.json())

    def assign_to_notebook(self, notebook_id: str, notebook_path: str) -> NotebookKernel:
        r_data = self._client.post(
            f"{self._path}{notebook_id}/notebook/kernel/", data={"path": notebook_path}
        )
        return NotebookKernel.from_server_data(r_data.json())

    def stop(self, notebook_id: str) -> None:
        self._client.delete(f"{self._path}{notebook_id}/kernels/{self.id}")



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Dict, List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.notebooks.enums import NotebookType, RunType, ScheduledRunStatus

scheduled_job_param_trafaret = t.Dict({t.Key("name"): t.String, t.Key("value"): t.String})


scheduled_job_payload_trafaret = t.Dict(
    {
        t.Key("uid"): t.String,
        t.Key("org_id"): t.String,
        t.Key("use_case_id"): t.String,
        t.Key("notebook_id"): t.String,
        t.Key("notebook_name"): t.String,
        t.Key("run_type"): t.Enum(*list(RunType)),
        t.Key("notebook_type"): t.Enum(*list(NotebookType)),
        t.Key("parameters"): t.List(scheduled_job_param_trafaret),
        t.Key("notebook_path", optional=True): t.String,
        t.Key("use_case_name", optional=True): t.String,
    }
).ignore_extra("*")

scheduled_run_revision_metadata_trafaret = t.Dict(
    {
        t.Key("id", optional=True): t.String,
        t.Key("name", optional=True): t.String,
    }
)


# TODO: [NB-4787] We are using trafaret's "ignore_extra" very liberally and this is a subset of properties
notebook_scheduled_run_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("use_case_id"): t.String,
        t.Key("status"): t.Enum(*list(ScheduledRunStatus)),
        t.Key("payload"): scheduled_job_payload_trafaret,
        t.Key("title", optional=True): t.String,
        t.Key("start_time", optional=True): t.String,
        t.Key("end_time", optional=True): t.String,
        t.Key("revision", optional=True): scheduled_run_revision_metadata_trafaret,
        t.Key("duration", optional=True): t.Int,
        t.Key("run_type", optional=True): t.Enum(*list(RunType)),
        t.Key("notebook_type", optional=True): t.Enum(*list(NotebookType)),
    }
).ignore_extra("*")


class ScheduledJobParam(APIObject):
    """
    DataRobot Schedule Job Parameter.

    Attributes
    ----------

    name : str
        The name of the parameter.
    value : str
        The value of the parameter.
    """

    _converter = scheduled_job_param_trafaret

    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value


class ScheduledJobPayload(APIObject):
    """
    DataRobot Schedule Job Payload.

    Attributes
    ----------

    uid : str
        The ID of the user who created the notebook schedule.
    org_id : str
        The ID of the user's organization who created the notebook schedule.
    use_case_id : str
        The ID of the Use Case that the notebook belongs to.
    notebook_id : str
        The ID of the notebook being run on a schedule.
    notebook_name : str
        The name of the notebook being run on a schedule.
    run_type : RunType
        The type of the run - either manual (triggered via UI or API) or scheduled.
    notebook_type: NotebookType
        The type of the notebook - either plain or codespace.
    parameters : List[ScheduledJobParam]
        The parameters being used in the notebook schedule. Can be an empty list.
    notebook_path : Optional[str]
        The path of the notebook to execute within the codespace. Optional. Required if notebook is in a codespace.
    use_case_name : Optional[str]
        The name of the Use Case that the notebook belongs to.
    """

    _converter = scheduled_job_payload_trafaret

    def __init__(
        self,
        uid: str,
        org_id: str,
        use_case_id: str,
        notebook_id: str,
        notebook_name: str,
        run_type: RunType,
        notebook_type: NotebookType,
        parameters: List[Dict[str, str]],
        notebook_path: Optional[str] = None,
        use_case_name: Optional[str] = None,
    ):
        self.uid = uid
        self.org_id = org_id
        self.use_case_id = use_case_id
        self.notebook_id = notebook_id
        self.notebook_name = notebook_name
        self.run_type = run_type
        self.notebook_type = notebook_type
        self.parameters = [ScheduledJobParam.from_server_data(param) for param in parameters]
        self.notebook_path = notebook_path
        self.use_case_name = use_case_name


class ScheduledRunRevisionMetadata(APIObject):
    """
    DataRobot Notebook Revision Metadata specifically for a scheduled run.

    Both id and name can be null if for example the job is still running or has failed.

    Attributes
    ----------

    id : Optional[str]
        The ID of the Notebook Revision. Optional.
    name : Optional[str]
        The name of the Notebook Revision. Optional.
    """

    _converter = scheduled_run_revision_metadata_trafaret

    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
    ):
        self.id = id
        self.name = name


class NotebookScheduledRun(APIObject):
    """
    DataRobot Notebook Scheduled Run. A historical run of a notebook schedule.

    Attributes
    ----------

    id : str
        The ID of the scheduled notebook job.
    use_case_id : str
        The Use Case ID of the scheduled notebook job.
    status : str
        The status of the run.
    payload : ScheduledJobPayload
        The payload used for the background job.
    title : Optional[str]
        The title of the job. Optional.
    start_time : Optional[str]
        The start time of the job. Optional.
    end_time : Optional[str]
        The end time of the job. Optional.
    revision : ScheduledRunRevisionMetadata
        Notebook revision data - ID and name.
    duration : Optional[int]
        The job duration in seconds. May be None for example while the job is running. Optional.
    run_type : Optional[RunType]
        The type of the run - either manual (triggered via UI or API) or scheduled. Optional.
    notebook_type: Optional[NotebookType]
        The type of the notebook - either plain or codespace. Optional.
    """

    _converter = notebook_scheduled_run_trafaret

    def __init__(
        self,
        id: str,
        use_case_id: str,
        status: ScheduledRunStatus,
        payload: Dict[str, Union[str, List[Dict[str, str]]]],
        title: Optional[str] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        revision: Optional[Dict[str, Optional[str]]] = None,
        duration: Optional[int] = None,
        run_type: Optional[RunType] = None,
        notebook_type: Optional[NotebookType] = None,
    ):
        self.id = id
        self.use_case_id = use_case_id
        self.status = status
        self.payload = ScheduledJobPayload.from_server_data(payload)
        self.title = title
        self.start_time = start_time
        self.end_time = end_time
        self.revision = (
            ScheduledRunRevisionMetadata.from_server_data(revision) if revision else None
        )
        self.duration = duration
        self.run_type = run_type
        self.notebook_type = notebook_type



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from copy import deepcopy
from pprint import pformat
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union, cast

import trafaret as t

from datarobot.enums import GridSearchAlgorithm, GridSearchSearchType, enum_to_list
from datarobot.models.api_object import APIObject

if TYPE_CHECKING:
    from datarobot.models.model import AdvancedTuningParamsType, Model, TuningParametersType
    from datarobot.models.modeljob import ModelJob

    class AugmentedTuningParametersType(TuningParametersType):
        value: Union[None, int, float, str, List[str]]


class NoParametersFoundException(Exception):
    """No parameters were found that matched the specified filter"""


class NonUniqueParametersException(Exception):
    """Multiple parameters were found that matched the specified filter"""

    def __init__(self, keys: Dict[str, Any], matching_params: List[TuningParametersType]) -> None:
        """Construct a NonUniqueParametersException

        Params
        ------
        keys : dict
            Keys used for filtering
        matching_params : list(dict)
            Parameters that matched the specified filter
        """
        self.matching_params = matching_params

        # Strip non-identifying fields from params for display purposes,
        # to make the error message more concise
        non_identifying_fields = {"defaultValue", "currentValue", "constraints"}
        filtered_params = [
            {key: value for key, value in param.items() if key not in non_identifying_fields}
            for param in matching_params
        ]

        message = dedent(
            """\
            Multiple matching parameters found for the specified keys.
            Keys: {}
            Parameters:
            {}"""
        ).format(pformat(keys), pformat(filtered_params))
        super().__init__(message)


class GridSearchArguments(APIObject):
    """
    Grid search arguments

    Attributes
    ----------
    search_type : GridSearchSearchType
        The type of grid search to be performed. If not specified, DataRobot performs Smart Search.
    algorithm : GridSearchAlgorithm (optional)
        The algorithm to apply when running the grid search.
        This is only applicable if the search type is specified and the search determines which algorithm to use.
        The following are the valid combinations of search type and algorithm:
        ------------------------------------------------------------
        | GridSearchSearchType.SMART | GridSearchAlgorithm.PATTERN_SEARCH (default) |
        | GridSearchSearchType.SMART | GridSearchAlgorithm.ACCELERATED_SEARCH |
        | GridSearchSearchType.BAYESIAN | GridSearchAlgorithm.TPE_SEARCH (default) |
        | GridSearchSearchType.BAYESIAN | GridSearchAlgorithm.GAUSSIAN_SEARCH |
        | GridSearchSearchType.BRUTE_FORCE | GridSearchAlgorithm.EXHAUSTIVE_SEARCH (default) |
        | GridSearchSearchType.BRUTE_FORCE | GridSearchAlgorithm.GREEDY_EXHAUSTIVE_SEARCH |
        ------------------------------------------------------------
    batch_size : int (optional)
        The number of iterations to perform in each batch.
    max_iterations : int (optional)
        Sets the maximum number of iterations to perform.
    random_state : int (optional)
        The random state/seed used for the grid search.
    wall_clock_time_limit : int (optional)
       The wall clock time limit, in seconds. The model with the best score, at this point, is selected.
    """

    _converter = t.Dict(
        {
            t.Key("grid_search_arguments", optional=True): t.List(
                t.Dict(
                    {
                        t.Key("search_type", optional=True): t.Enum(
                            *enum_to_list(GridSearchSearchType)
                        ),
                        t.Key("algorithm", optional=True): t.Enum(
                            *enum_to_list(GridSearchAlgorithm)
                        ),
                        t.Key("batch_size", optional=True): t.Int(),
                        t.Key("max_iterations", optional=True): t.Int(),
                        t.Key("random_state", optional=True): t.Int(),
                        t.Key("wall_clock_time_limit", optional=True): t.Int(),
                    }
                )
            ),
        }
    ).allow_extra("*")

    def __init__(
        self,
        search_type: Optional[GridSearchSearchType] = None,
        algorithm: Optional[GridSearchAlgorithm] = None,
        batch_size: Optional[int] = None,
        max_iterations: Optional[int] = None,
        random_state: Optional[int] = None,
        wall_clock_time_limit: Optional[int] = None,
    ) -> None:
        self.search_type = search_type if search_type else GridSearchSearchType.SMART
        self.algorithm = algorithm
        self.batch_size = batch_size
        self.max_iterations = max_iterations
        self.random_state = random_state
        self.wall_clock_time_limit = wall_clock_time_limit

    def to_api_payload(self) -> Dict[str, Any]:
        """Convert the GridSearchArguments to an API payload"""
        payload: Dict[str, Any] = {}
        if self.search_type:
            payload["searchType"] = self.search_type
        if self.algorithm:
            payload["algorithm"] = self.algorithm
        if self.batch_size:
            payload["batchSize"] = self.batch_size
        if self.max_iterations:
            payload["maxIterations"] = self.max_iterations
        if self.random_state:
            payload["randomState"] = self.random_state
        if self.wall_clock_time_limit:
            payload["wallClockTimeLimit"] = self.wall_clock_time_limit
        return payload


class AdvancedTuningSession:
    """A session enabling users to configure and run advanced tuning for a model.

    Every model contains a set of one or more tasks.  Every task contains a set of
    zero or more parameters.  This class allows tuning the values of each parameter
    on each task of a model, before running that model.

    This session is client-side only and is not persistent.
    Only the final model, constructed when `run` is called, is persisted on the DataRobot server.

    Attributes
    ----------
    description : str
        Description for the new advance-tuned model.
        Defaults to the same description as the base model.
    """

    def __init__(
        self, model: Model, grid_search_arguments: Optional[GridSearchArguments] = None
    ) -> None:
        """Initiate an Advanced Tuning session.

        Params
        ------
        model : datarobot.models.model.Model
        grid_search_arguments : datarobot.models.advanced_tuning.GridSearchArguments
            Grid search arguments
        """
        self._new_values: Dict[str, Union[int, float, str, List[str]]] = {}
        self._grid_search_arguments = grid_search_arguments
        self._model = model

        param_info = model.get_advanced_tuning_parameters()
        self._available_params = param_info["tuning_parameters"]
        self.description = param_info.get("tuning_description")

    def _get_parameter_id(
        self,
        task_name: Optional[str] = None,
        parameter_name: Optional[str] = None,
        parameter_id: Optional[str] = None,
    ) -> str:
        """Return the ID of the one parameter that matches the specified fields.

        Returns
        -------
        str

        Raises
        ------
        NoParametersFoundException
            if no matching parameters are found.
        NonUniqueParametersException
            if multiple parameters matched the specified filtering criteria
        """
        filtered_params = (x for x in self._available_params)
        if parameter_id:
            # Should be unique but filter normally just in case it's ever not unique
            filtered_params = (x for x in filtered_params if x["parameter_id"] == parameter_id)
        if parameter_name:
            filtered_params = (x for x in filtered_params if x["parameter_name"] == parameter_name)
        if task_name:
            filtered_params = (x for x in filtered_params if x["task_name"] == task_name)

        filtered_params_list = list(filtered_params)

        if len(filtered_params_list) == 0:
            raise NoParametersFoundException(
                "No parameters found with task_name of {} and parameter_name of {}".format(
                    repr(task_name) if task_name else "(unspecified)",
                    repr(parameter_name) if parameter_name else "(unspecified)",
                )
            )

        if len(filtered_params_list) > 1:
            key = {}
            if task_name:
                key["task_name"] = task_name
            if parameter_name:
                key["parameter_name"] = parameter_name
            if parameter_id:
                key["parameter_id"] = parameter_id
            raise NonUniqueParametersException(key, filtered_params_list)

        return filtered_params_list[0]["parameter_id"]

    def get_task_names(self) -> List[str]:
        """Get the list of task names that are available for this model

        Returns
        -------
        list(str)
            List of task names
        """
        return sorted({x["task_name"] for x in self._available_params})

    def get_parameter_names(self, task_name: str) -> List[str]:
        """Get the list of parameter names available for a specific task

        Returns
        -------
        list(str)
            List of parameter names
        """
        return [x["parameter_name"] for x in self._available_params if x["task_name"] == task_name]

    def set_parameter(
        self,
        value: Union[int, float, str, List[str]],
        task_name: Optional[str] = None,
        parameter_name: Optional[str] = None,
        parameter_id: Optional[str] = None,
    ) -> None:
        """Set the value of a parameter to be used

        The caller must supply enough of the optional arguments to this function
        to uniquely identify the parameter that is being set.
        For example, a less-common parameter name such as
        'building_block__complementary_error_function' might only be used once (if at all)
        by a single task in a model.  In which case it may be sufficient to simply specify
        'parameter_name'.  But a more-common name such as 'random_seed' might be used by
        several of the model's tasks, and it may be necessary to also specify 'task_name'
        to clarify which task's random seed is to be set.
        This function only affects client-side state. It will not check that the new parameter
        value(s) are valid.

        Parameters
        ----------
        task_name : str
            Name of the task whose parameter needs to be set
        parameter_name : str
            Name of the parameter to set
        parameter_id : str
            ID of the parameter to set
        value : int, float, list, or str
            New value for the parameter, with legal values determined by the parameter being set

        Raises
        ------
        NoParametersFoundException
            if no matching parameters are found.
        NonUniqueParametersException
            if multiple parameters matched the specified filtering criteria
        """
        parameter_id = self._get_parameter_id(
            task_name=task_name, parameter_name=parameter_name, parameter_id=parameter_id
        )

        self._new_values[parameter_id] = value

    def _add_value_to_param(self, param: TuningParametersType) -> AugmentedTuningParametersType:
        """Given a 'param' dict, add a new user-specified value (if any) and return"""
        new_param = cast("AugmentedTuningParametersType", deepcopy(param))
        new_param["value"] = self._new_values.get(param["parameter_id"])
        return new_param

    def get_parameters(self) -> AdvancedTuningParamsType:
        """Returns the set of parameters available to this model

        The returned parameters have one additional key, "value", reflecting any new values that
        have been set in this AdvancedTuningSession.  When the session is run, "value" will be used,
        or if it is unset, "current_value".


        Returns
        -------
        parameters : dict
            "Parameters" dictionary, same as specified on `Model.get_advanced_tuning_params`.

        An additional field is added per parameter to the 'tuning_parameters' list in the dictionary:

        value : int, float, list, or str
            The current value of the parameter.  `None` if none has been specified.
        """

        return {
            "tuning_description": self.description,
            "tuning_parameters": [
                self._add_value_to_param(param) for param in self._available_params
            ],
        }

    def run(self) -> ModelJob:
        """Submit this model for Advanced Tuning.

        Returns
        -------
        datarobot.models.modeljob.ModelJob
            The created job to build the model
        """
        return self._model.advanced_tune(
            self._new_values, self.description, self._grid_search_arguments
        )



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
import json
from typing import Any, Dict, List, Optional, Union

import trafaret as t
from trafaret.contrib.rfc_3339 import DateTime

from datarobot._compat import String
from datarobot.enums import CredentialTypes
from datarobot.models.api_object import APIObject
from datarobot.utils import pagination, rawdict


class Credential(APIObject):  # pylint: disable=missing-class-docstring
    _path = "credentials/"
    _converter = t.Dict(
        {
            t.Key("name"): String(),
            t.Key("credential_id"): String(),
            t.Key("creation_date"): DateTime(),
            t.Key("credential_type"): String(),
            t.Key("description"): String(allow_blank=True),
        }
    ).allow_extra("*")

    def __init__(
        self,
        credential_id: Optional[str] = None,
        name: Optional[str] = None,
        credential_type: Optional[str] = None,
        creation_date: Optional[datetime] = None,
        description: Optional[str] = None,
    ) -> None:
        self.credential_id = credential_id
        self.name = name
        self.credential_type = credential_type
        self.creation_date = creation_date
        self.description = description

    @classmethod
    def list(cls) -> List[Credential]:
        """
        Returns list of available credentials.

        Returns
        -------
        credentials : list of Credential instances
            contains a list of available credentials.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_sources = dr.Credential.list()
            >>> data_sources
            [
                Credential('5e429d6ecf8a5f36c5693e03', 'my_s3_cred', 's3'),
                Credential('5e42cc4dcf8a5f3256865840', 'my_jdbc_cred', 'jdbc'),
            ]
        """

        return [
            cls.from_server_data(item) for item in pagination.unpaginate(cls._path, {}, cls._client)
        ]

    @classmethod
    def get(cls, credential_id: str) -> Credential:
        """
        Gets the Credential.

        Parameters
        ----------
        credential_id : str
            the identifier of the credential.

        Returns
        -------
        credential : Credential
            the requested credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.get('5a8ac9ab07a57a0001be501f')
            >>> cred
            Credential('5e429d6ecf8a5f36c5693e03', 'my_s3_cred', 's3'),
        """
        return cls.from_location(f"{cls._path}{credential_id}/")

    def delete(self) -> None:
        """
        Deletes the Credential the store.

        Parameters
        ----------
        credential_id : str
            the identifier of the credential.

        Returns
        -------
        credential : Credential
            the requested credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.get('5a8ac9ab07a57a0001be501f')
            >>> cred.delete()
        """
        self._client.delete(f"{self._path}{self.credential_id}/")

    @classmethod
    def create_basic(
        cls,
        name: str,
        user: str,
        password: str,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the credentials.

        Parameters
        ----------
        name : str
            the name to use for this set of credentials.
        user : str
            the username to store for this set of credentials.
        password : str
            the password to store for this set of credentials.
        description : Optional[str]
            the description to use for this set of credentials.

        Returns
        -------
        credential : Credential
            the created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_basic(
            ...     name='my_basic_cred',
            ...     user='username',
            ...     password='password',
            ... )
            >>> cred
            Credential('5e429d6ecf8a5f36c5693e03', 'my_basic_cred', 'basic'),
        """
        payload = {
            "name": name,
            "credentialType": CredentialTypes.BASIC.value,
            "user": user,
            "password": password,
            "description": description,
        }
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    @classmethod
    def create_oauth(
        cls,
        name: str,
        token: str,
        refresh_token: str,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the OAUTH credentials.

        Parameters
        ----------
        name : str
            the name to use for this set of credentials.
        token: str
            the OAUTH token
        refresh_token: str
            The OAUTH token
        description : Optional[str]
            the description to use for this set of credentials.

        Returns
        -------
        credential : Credential
            the created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_oauth(
            ...     name='my_oauth_cred',
            ...     token='XXX',
            ...     refresh_token='YYY',
            ... )
            >>> cred
            Credential('5e429d6ecf8a5f36c5693e03', 'my_oauth_cred', 'oauth'),
        """
        payload = {
            "name": name,
            "credentialType": CredentialTypes.OAUTH.value,
            "token": token,
            "refreshToken": refresh_token,
            "description": description,
        }
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    @classmethod
    def create_s3(
        cls,
        name: str,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        config_id: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the S3 credentials.

        Parameters
        ----------
        name : str
            the name to use for this set of credentials.
        aws_access_key_id : Optional[str]
            the AWS access key id.
        aws_secret_access_key : Optional[str]
            the AWS secret access key.
        aws_session_token : Optional[str]
            the AWS session token.
        config_id: Optional[str]
            The ID of the saved shared secure configuration. If specified, cannot include awsAccessKeyId,
            awsSecretAccessKey or awsSessionToken.
        description : Optional[str]
            the description to use for this set of credentials.

        Returns
        -------
        credential : Credential
            the created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_s3(
            ...     name='my_s3_cred',
            ...     aws_access_key_id='XXX',
            ...     aws_secret_access_key='YYY',
            ...     aws_session_token='ZZZ',
            ... )
            >>> cred
            Credential('5e429d6ecf8a5f36c5693e03', 'my_s3_cred', 's3'),
        """
        payload = {
            "name": name,
            "credentialType": CredentialTypes.S3.value,
            "awsAccessKeyId": aws_access_key_id,
            "awsSecretAccessKey": aws_secret_access_key,
            "awsSessionToken": aws_session_token,
            "configId": config_id,
            "description": description,
        }
        payload = {key: value for key, value in payload.items() if value is not None}
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    @classmethod
    def create_azure(
        cls,
        name: str,
        azure_connection_string: str,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the Azure storage credentials.

        Parameters
        ----------
        name : str
            the name to use for this set of credentials.
        azure_connection_string : str
            the Azure connection string.
        description : Optional[str]
            the description to use for this set of credentials.

        Returns
        -------
        credential : Credential
            the created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_azure(
            ...     name='my_azure_cred',
            ...     azure_connection_string='XXX',
            ... )
            >>> cred
            Credential('5e429d6ecf8a5f36c5693e03', 'my_azure_cred', 'azure'),
        """
        payload = {
            "name": name,
            "credentialType": CredentialTypes.AZURE.value,
            "azureConnectionString": azure_connection_string,
            "description": description,
        }
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    @classmethod
    def create_snowflake_key_pair(
        cls,
        name: str,
        user: Optional[str] = None,
        private_key: Optional[str] = None,
        passphrase: Optional[str] = None,
        config_id: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the Snowflake Key Pair credentials.

        Parameters
        ----------
        name : str
            the name to use for this set of credentials.
        user: Optional[str]
            the Snowflake login name
        private_key: Optional[str]
            the private key copied exactly from user private key file. Since it contains multiple
            lines, when assign to a variable, put the key string inside triple-quotes
        passphrase: Optional[str]
            the string used to encrypt the private key
        config_id: Optional[str]
            The ID of the saved shared secure configuration. If specified, cannot include user,
            privateKeyStr or passphrase.
        description : Optional[str]
            the description to use for this set of credentials.

        Returns
        -------
        credential : Credential
            the created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_snowflake_key_pair(
            ...     name='key_pair_cred',
            ...     user='XXX',
            ...     private_key='YYY',
            ...     passphrase='ZZZ',
            ... )
            >>> cred
            Credential('5e429d6ecf8a5f36c5693e03', 'key_pair_cred', 'snowflake_key_pair_user_account'),
        """
        payload = {
            "name": name,
            "credentialType": CredentialTypes.SNOWFLAKE_KEY_PAIR_AUTH.value,
            "user": user,
            "privateKeyStr": private_key,
            "passphrase": passphrase,
            "configId": config_id,
            "description": description,
        }
        payload = {key: value for key, value in payload.items() if value is not None}
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    @classmethod
    def create_databricks_access_token(
        cls,
        name: str,
        databricks_access_token: str,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the Databricks access token credentials.

        Parameters
        ----------
        name : str
            the name to use for this set of credentials.
        databricks_access_token: Optional[str]
            the Databricks personal access token
        description : Optional[str]
            the description to use for this set of credentials.

        Returns
        -------
        credential : Credential
            the created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_databricks_access_token(
            ...     name='access_token_cred',
            ...     databricks_access_token='XXX',
            ... )
            >>> cred
            Credential('5e429d6ecf8a5f36c5693e03', 'access_token_cred', 'databricks_access_token_account'),
        """
        payload = {
            "name": name,
            "credentialType": CredentialTypes.DATABRICKS_ACCESS_TOKEN.value,
            "databricksAccessToken": databricks_access_token,
            "description": description,
        }
        payload = {key: value for key, value in payload.items() if value is not None}
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    @classmethod
    def create_databricks_service_principal(
        cls,
        name: str,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        config_id: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the Databricks access token credentials.

        Parameters
        ----------
        name : str
            the name to use for this set of credentials.
        client_id: Optional[str]
            the client ID for Databricks Service Principal
        client_secret: Optional[str]
            the client secret for Databricks Service Principal
        config_id: Optional[str]
            The ID of the saved shared secure configuration. If specified, cannot include clientId
            and clientSecret.
        description : Optional[str]
            the description to use for this set of credentials.

        Returns
        -------
        credential : Credential
            the created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_databricks_service_principal(
            ...     name='svc_principal_cred',
            ...     client_id='XXX',
            ...     client_secret='XXX',
            ... )
            >>> cred
            Credential('5e429d6ecf8a5f36c5693e03', 'svc_principal_cred', 'databricks_service_principal_account'),
        """
        payload = {
            "name": name,
            "credentialType": CredentialTypes.DATABRICKS_SERVICE_PRINCIPAL.value,
            "clientId": client_id,
            "clientSecret": client_secret,
            "configId": config_id,
            "description": description,
        }
        payload = {key: value for key, value in payload.items() if value is not None}
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    @classmethod
    def create_azure_service_principal(
        cls,
        name: str,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        azure_tenant_id: Optional[str] = None,
        config_id: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the Azure service principal credentials.

        Parameters
        ----------
        name : str
            The name to use for these credentials.
        client_id : Optional[str]
            The client ID.
        client_secret : Optional[str]
            The client secret.
        azure_tenant_id : Optional[str]
            The Azure tenant ID.
        config_id: Optional[str]
            The ID of the saved secure configuration. If specified, the ID cannot include `clientId`,
            `clientSecret`, or `azureTenantId`.
        description : Optional[str]
            The description to use for these credentials.

        Returns
        -------
        credential : Credential
            the created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_azure_service_principal(
            ...     name='my_azure_service_principal_cred',
            ...     client_id='XXX',
            ...     client_secret='YYY',
            ...     azure_tenant_id='ZZZ',
            ... )
            >>> cred
            Credential('66c9172d8b7a361cda126f5c', 'my_azure_service_principal_cred', 'azure_service_principal')
        """
        payload = {
            "name": name,
            "credentialType": CredentialTypes.AZURE_SERVICE_PRINCIPAL.value,
            "clientId": client_id,
            "clientSecret": client_secret,
            "azureTenantId": azure_tenant_id,
            "configId": config_id,
            "description": description,
        }
        payload = {key: value for key, value in payload.items() if value is not None}
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    @classmethod
    def create_adls_oauth(
        cls,
        name: str,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        oauth_scopes: Optional[List[str]] = None,
        config_id: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the ADLS OAuth credentials.

        Parameters
        ----------
        name : str
            The name to use for these credentials.
        client_id : Optional[str]
            The client ID.
        client_secret : Optional[str]
            The client secret.
        oauth_scopes : List[str], optional
            The OAuth scopes.
        config_id: Optional[str]
            The ID of the saved shared secure configuration. If specified, cannot include clientId,
            `clientSecret`, or `oauthScopes`.
        description : Optional[str]
            The description to use for the ADLS OAuth credentials.

        Returns
        -------
        credential : Credential
            The created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_adls_oauth(
            ...     name='my_adls_oauth_cred',
            ...     client_id='XXX',
            ...     client_secret='YYY',
            ...     oauth_scopes=['ZZZ'],
            ... )
            >>> cred
            Credential('66c91e0f03010d4790735220', 'my_adls_oauth_cred', 'adls_gen2_oauth')
        """
        payload = {
            "name": name,
            "credentialType": CredentialTypes.ADLS_OAUTH.value,
            "clientId": client_id,
            "clientSecret": client_secret,
            "oauthScopes": oauth_scopes,
            "configId": config_id,
            "description": description,
        }
        payload = {key: value for key, value in payload.items() if value is not None}
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    def __repr__(self) -> str:
        return "{}('{}', '{}', '{}')".format(
            self.__class__.__name__,
            self.credential_id,
            self.name,
            self.credential_type,
        )

    @classmethod
    def create_gcp(
        cls,
        name: str,
        gcp_key: Optional[Union[str, Dict[str, str]]] = None,
        description: Optional[str] = None,
    ) -> Credential:
        """
        Creates the GCP credentials.

        Parameters
        ----------
        name : str
            the name to use for this set of credentials.
        gcp_key : str | dict
            the GCP key in json format or parsed as dict.
        description : Optional[str]
            the description to use for this set of credentials.

        Returns
        -------
        credential : Credential
            the created credential.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> cred = dr.Credential.create_gcp(
            ...     name='my_gcp_cred',
            ...     gcp_key='XXX',
            ... )
            >>> cred
            Credential('5e429d6ecf8a5f36c5693e03', 'my_gcp_cred', 'gcp'),
        """

        if isinstance(gcp_key, str):
            try:
                gcp_key = json.loads(gcp_key)
            except ValueError as e:
                raise ValueError(f"Could not parse gcp_key: {e}")

        payload = {
            "name": name,
            "credentialType": CredentialTypes.GCP.value,
            "gcpKey": rawdict(gcp_key),  # type: ignore[arg-type]
            "description": description,
        }
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    def update(
        self, name: Optional[str] = None, description: Optional[str] = None, **kwargs: Any
    ) -> None:
        """Update the credential values of an existing credential. Updates this object in place.

        .. versionadded:: v3.2

        Parameters
        ----------
        name : str
            The name to use for this set of credentials.
        description : Optional[str]
            The description to use for this set of credentials; if omitted, and name is not
            omitted, then it clears any previous description for that name.
        kwargs : Keyword arguments specific to the given credential_type that should be updated.
        """
        if name is not None:
            kwargs["name"] = name
        if description is not None:
            kwargs["description"] = description

        self._client.patch(f"{self._path}{self.credential_id}/", data=kwargs)

        if name is not None:
            self.name = name
        if description is not None:
            self.description = description


BasicCredentialsSchema = t.Dict(
    {
        t.Key("credentialType"): t.Atom("basic"),
        t.Key("user"): String(),
        t.Key("password"): String(),
    }
).allow_extra("*")

S3CredentialsSchema = t.Dict(
    {
        t.Key("credentialType"): t.Atom("s3"),
        t.Key("awsAccessKeyId", optional=True): String(),
        t.Key("awsSecretAccessKey", optional=True): String(),
        t.Key("awsSessionToken", optional=True): String(),
        t.Key("configId", optional=True): String(),
    }
).allow_extra("*")

OauthCredentialsSchema = t.Dict(
    {
        t.Key("credentialType"): t.Atom("oauth"),
        t.Key("oauthRefreshToken"): String(),
        t.Key("oauthClientId", optional=True): String(),
        t.Key("oauthClientSecret", optional=True): String(),
        t.Key("oauthAccessToken", optional=True): String(),
    }
).allow_extra("*")

SnowflakeKeyPairCredentialsSchema = t.Dict(
    {
        t.Key("credentialType"): t.Atom("snowflake_key_pair_user_account"),
        t.Key("user", optional=True): t.Or(String, t.Null),
        t.Key("privateKeyStr", optional=True): t.Or(String, t.Null),
        t.Key("passphrase", optional=True): t.Or(String, t.Null),
        t.Key("configId", optional=True): t.Or(String, t.Null),
    }
).allow_extra("*")

DatabricksAccessTokenCredentialsSchema = t.Dict(
    {
        t.Key("credentialType"): t.Atom("databricks_access_token_account"),
        t.Key("databricksAccessToken"): String(),
    }
).allow_extra("*")

DatabricksServicePrincipalCredentialsSchema = t.Dict(
    {
        t.Key("credentialType"): t.Atom("databricks_service_principal_account"),
        t.Key("clientId", optional=True): String(),
        t.Key("clientSecret", optional=True): String(),
        t.Key("configId", optional=True): String(),
    }
).allow_extra("*")

AzureServicePrincipalCredentialsSchema = t.Dict(
    {
        t.Key("credentialType"): t.Atom("azure_service_principal"),
        t.Key("clientId", optional=True): String(),
        t.Key("clientSecret", optional=True): String(),
        t.Key("azureTenantId", optional=True): String(),
        t.Key("configId", optional=True): String(),
    }
).allow_extra("*")

AdlsOAuthCredentialsSchema = t.Dict(
    {
        t.Key("credentialType"): t.Atom("adls_gen2_oauth"),
        t.Key("clientId", optional=True): String(),
        t.Key("clientSecret", optional=True): String(),
        t.Key("oauthScopes", optional=True): t.List(String()),
        t.Key("configId", optional=True): String(),
    }
).allow_extra("*")

AnyCredentialsSchema = t.Dict({t.Key("credentialType"): String()}).allow_extra("*")

CredentialDataSchema = t.Or(
    BasicCredentialsSchema,
    S3CredentialsSchema,
    OauthCredentialsSchema,
    SnowflakeKeyPairCredentialsSchema,
    DatabricksAccessTokenCredentialsSchema,
    DatabricksServicePrincipalCredentialsSchema,
    AzureServicePrincipalCredentialsSchema,
    AdlsOAuthCredentialsSchema,
    AnyCredentialsSchema,
)



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import trafaret as t

from datarobot.models.api_object import APIObject

notebook_settings_trafaret = t.Dict(
    {
        t.Key("show_line_numbers"): t.Bool,
        t.Key("hide_cell_titles"): t.Bool,
        t.Key("hide_cell_outputs"): t.Bool,
        t.Key("show_scrollers"): t.Bool,
        t.Key("hide_cell_footers"): t.Bool,
        t.Key("highlight_whitespace"): t.Bool,
    }
).ignore_extra("*")


class NotebookSettings(APIObject):
    """
    Settings for a DataRobot Notebook.

    Attributes
    ----------

    show_line_numbers : bool
        Whether line numbers in cells should be displayed.
    hide_cell_titles : bool
        Whether cell titles should be displayed.
    hide_cell_outputs : bool
        Whether the cell outputs should be displayed.
    show_scrollers : bool
        Whether scroll bars should be shown on cells.
    hide_cell_footers : bool
        Whether footers should be shown on cells.
    highlight_whitespace : bool
        Whether whitespace should be highlighted or not.
    """

    _converter = notebook_settings_trafaret

    def __init__(
        self,
        show_line_numbers: bool,
        hide_cell_titles: bool,
        hide_cell_outputs: bool,
        show_scrollers: bool,
        hide_cell_footers: bool,
        highlight_whitespace: bool,
    ):
        self.show_line_numbers = show_line_numbers
        self.hide_cell_titles = hide_cell_titles
        self.hide_cell_outputs = hide_cell_outputs
        self.show_scrollers = show_scrollers
        self.hide_cell_footers = hide_cell_footers
        self.highlight_whitespace = highlight_whitespace



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional

import trafaret as t
from typing_extensions import TypedDict

from datarobot.models.api_object import APIObject
from datarobot.models.custom_application import CustomApplication
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


class CustomApplicationSourceResources(TypedDict):
    """Resource configuration for a Custom Application Source."""

    replicas: int
    resource_label: str
    service_web_requests_on_root_path: Optional[bool]
    session_affinity: Optional[bool]


class CustomApplicationSourceVersion(TypedDict):
    """Version information for a Custom Application Source."""

    id: str
    label: Optional[str]
    created_at: str
    updated_at: str
    is_frozen: bool


# Trafaret validator for resources field (can be dict or null)
_resources_validator = t.Or(
    t.Dict(
        {
            t.Key("replicas"): t.Int,
            t.Key("resource_label"): t.String,
            t.Key("service_web_requests_on_root_path", optional=True): t.Bool,
            t.Key("session_affinity", optional=True): t.Bool,
        }
    ).ignore_extra("*"),
    t.Null,
)

# Trafaret validator for latest_version field
# The latest_version can contain resources nested within it
_latest_version_validator = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("label", optional=True): t.Or(t.String, t.Null),
        t.Key("created_at"): t.String,
        t.Key("updated_at"): t.String,
        t.Key("is_frozen"): t.Bool,
        t.Key("resources", optional=True): _resources_validator,
    }
).ignore_extra("*")


class CustomApplicationSource(APIObject):
    """
    A DataRobot custom application source that serves as a template for creating custom applications.

    Custom application sources define the code, configuration, and resource requirements
    used when creating custom applications. They can have resource configurations
    that determine the default resource allocation for applications created from this source.

    Attributes
    ----------
    id : str
        The ID of the custom application source.
    name : str
        The name of the custom application source.
    latest_version : CustomApplicationSourceVersion
        Information about the latest version of this source.
    user_id : str
        The ID of the user who created the source.
    org_id : str
        The organization ID.
    permissions : List[str]
        List of permissions for the current user.
    created_at : str
        Timestamp when the source was created.
    updated_at : str
        Timestamp when the source was last updated.
    created_by : str
        Email of the user who created the source.
    creator_first_name : str
        First name of the creator.
    creator_last_name : str
        Last name of the creator.
    creator_userhash : str
        Userhash of the creator.

    Notes
    -----
    Resource configuration is stored in the latest_version field and can be accessed
    via the get_resources() method.
    """

    _path = "customApplicationSources/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("name"): t.String,
            t.Key("latest_version"): _latest_version_validator,
            t.Key("user_id"): t.String,
            t.Key("org_id"): t.String,
            t.Key("permissions"): t.List(t.String),
            t.Key("created_at"): t.String,
            t.Key("updated_at"): t.String,
            t.Key("created_by"): t.String,
            t.Key("creator_first_name"): t.String,
            t.Key("creator_last_name"): t.String,
            t.Key("creator_userhash"): t.String,
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        latest_version: Dict[str, Any],
        user_id: str,
        org_id: str,
        permissions: List[str],
        created_at: str,
        updated_at: str,
        created_by: str,
        creator_first_name: str,
        creator_last_name: str,
        creator_userhash: str,
    ):
        self.id = id
        self.name = name
        self.latest_version = latest_version
        self.user_id = user_id
        self.org_id = org_id
        self.permissions = permissions
        self.created_at = created_at
        self.updated_at = updated_at
        self.created_by = created_by
        self.creator_first_name = creator_first_name
        self.creator_last_name = creator_last_name
        self.creator_userhash = creator_userhash

    def __repr__(self) -> str:
        return f"CustomApplicationSource({self.name!r}, id={self.id!r})"

    @classmethod
    def list(
        cls,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[CustomApplicationSource]:
        """
        Retrieve a list of custom application sources.

        Parameters
        ----------
        offset : Optional[int]
            Optional. Retrieve sources in a list after this number.
        limit : Optional[int]
            Optional. Retrieve only this number of sources.

        Returns
        -------
        sources : List[CustomApplicationSource]
            The requested list of custom application sources.
        """
        query = {"offset": offset, "limit": limit}
        sources = unpaginate(initial_url=cls._path, initial_params=query, client=cls._client)
        return [cls.from_server_data(source) for source in sources]

    @classmethod
    def get(cls, source_id: str) -> CustomApplicationSource:
        """
        Retrieve a single custom application source with full details.

        Parameters
        ----------
        source_id : str
            The ID of the custom application source to retrieve.

        Returns
        -------
        source : CustomApplicationSource
            The requested custom application source.
        """
        r_data = cls._client.get(f"{cls._path}{source_id}/")
        return cls.from_server_data(r_data.json())

    @classmethod
    def create(cls, name: str) -> CustomApplicationSource:
        """
        Create a new custom application source.

        Parameters
        ----------
        name : str
            The name for the new custom application source.

        Returns
        -------
        source : CustomApplicationSource
            The newly created custom application source.
        """
        payload = {"name": name}
        r_data = cls._client.post(cls._path, data=payload)
        return cls.from_server_data(r_data.json())

    def update(self, name: Optional[str] = None) -> CustomApplicationSource:
        """
        Update this custom application source.

        Parameters
        ----------
        name : Optional[str]
            New name for the source. If None, name will not be changed.

        Returns
        -------
        source : CustomApplicationSource
            The updated custom application source.
        """
        payload = {}
        if name is not None:
            payload["name"] = name

        if not payload:
            return self  # No changes to make

        r_data = self._client.patch(f"{self._path}{self.id}/", data=payload)
        return self.from_server_data(r_data.json())

    def delete(self, hard_delete: bool = False) -> None:
        """
        Delete this custom application source.

        Parameters
        ----------
        hard_delete : bool, optional
            If True, permanently delete the source and all its data.
            If False (default), soft delete the source (can be recovered).

        Note
        ----
        Deleting a source will affect all its versions. Applications created from this
        source will continue to run but cannot be recreated from the source.

        Examples
        --------
        .. code-block:: python

            from datarobot import CustomApplicationSource

            source = CustomApplicationSource.get("source_id")
            source.delete()  # Soft delete
            source.delete(hard_delete=True)  # Permanent delete
        """
        params = {"hardDelete": hard_delete} if hard_delete else {}
        self._client.delete(f"{self._path}{self.id}/", params=params)

    def get_resources(self) -> Optional[Dict[str, Any]]:
        """
        Get resource configuration for applications created from this source.

        Returns
        -------
        resources : Optional[Dict[str, Any]]
            Resource configuration including replicas, resource bundle, and other settings.
            Returns None if no resources are configured.
        """
        # Resources are in the latest_version object
        if (
            self.latest_version is None
            or not isinstance(self.latest_version, dict)
            or "resources" not in self.latest_version
            or self.latest_version["resources"] is None
        ):
            return None

        resources = self.latest_version["resources"]

        # Resources are already in snake_case format from the API client conversion
        return {
            "replicas": resources.get("replicas"),
            "resource_label": resources.get("resource_label"),
            "service_web_requests_on_root_path": resources.get("service_web_requests_on_root_path"),
            "session_affinity": resources.get("session_affinity"),
        }

    def get_resource_summary(self) -> Optional[Dict[str, Any]]:
        """
        Get a human-readable summary of resource configuration.

        Returns
        -------
        summary : Optional[Dict[str, Any]]
            A summary of resource configuration with readable descriptions.
            Returns None if no resources are configured.
        """
        # Resources are in the latest_version object
        if (
            self.latest_version is None
            or not isinstance(self.latest_version, dict)
            or "resources" not in self.latest_version
            or self.latest_version["resources"] is None
        ):
            return None

        resources = self.latest_version["resources"]

        return {
            "resource_bundle": resources.get("resource_label"),
            "scaling": {
                "replicas": resources.get("replicas"),
                "session_affinity": resources.get("session_affinity"),
            },
            "networking": {
                "service_web_requests_on_root_path": resources.get(
                    "service_web_requests_on_root_path"
                ),
            },
        }

    @property
    def has_resources_configured(self) -> bool:
        """
        Check if this source has resource configuration.

        Returns
        -------
        bool
            True if resources are configured, False otherwise.
        """
        return (
            self.latest_version is not None
            and isinstance(self.latest_version, dict)
            and "resources" in self.latest_version
            and self.latest_version["resources"] is not None
        )

    def get_details(self) -> Dict[str, Any]:
        """
        Get comprehensive details about this custom application source.

        Returns
        -------
        details : Dict[str, Any]
            Comprehensive source details including metadata, version info, and resources.
        """
        return {
            "basic_info": {
                "id": self.id,
                "name": self.name,
                "created_at": self.created_at,
                "updated_at": self.updated_at,
            },
            "creator_info": {
                "created_by": self.created_by,
                "creator_name": f"{self.creator_first_name} {self.creator_last_name}",
                "creator_userhash": self.creator_userhash,
            },
            "latest_version": {
                "id": self.latest_version.get("id"),
                "label": self.latest_version.get("label"),
                "created_at": self.latest_version.get("created_at"),
                "updated_at": self.latest_version.get("updated_at"),
                "is_frozen": self.latest_version.get("is_frozen"),
            },
            "resources": self.get_resource_summary(),
            "permissions": self.permissions,
            "has_resources": self.has_resources_configured,
        }

    @classmethod
    def get_by_name(cls, name: str) -> Optional[CustomApplicationSource]:
        """
        Find a custom application source by name.

        Parameters
        ----------
        name : str
            The name of the custom application source to find.

        Returns
        -------
        source : Optional[CustomApplicationSource]
            The custom application source if found, None otherwise.
        """
        sources = cls.list()
        for source in sources:
            if source.name == name:
                return source
        return None

    def create_application(
        self,
        name: str,
        resources: Optional[Dict[str, Any]] = None,
        environment_id: Optional[str] = None,
    ) -> CustomApplication:
        """
        Create a Custom Application from this source.

        Parameters
        ----------
        name : str
            Name for the new application.
        resources : Optional[Dict[str, Any]]
            Resource configuration for the application. If None, uses source defaults.
        environment_id : Optional[str]
            The ID of the execution environment to use. If None, uses source default.
        external_access_enabled : bool
            Whether to enable external access. Default is False.
        external_access_recipients : Optional[List[str]]
            List of email addresses for external access recipients.

        Returns
        -------
        application : CustomApplication
            The newly created Custom Application.

        Examples
        --------
        .. code-block:: python

            from datarobot import CustomApplicationSource

            source = CustomApplicationSource.get("source_id")

            # Create with default settings
            app1 = source.create_application("My App")

            # Create with custom resources and environment
            app2 = source.create_application(
                "My App",
                resources={"replicas": 2, "resourceLabel": "cpu.large"},
                environment_id="env_id_123"
            )
        """
        # Build the payload - use applicationSourceId to create from source
        # The API requires ONLY ONE of: baseEnvironmentId, applicationSourceId, or applicationSourceVersionId
        payload: Dict[str, Any] = {
            "name": name,
            "applicationSourceId": self.id,
        }

        # Add optional environment ID
        if environment_id:
            payload["envVersionId"] = environment_id

        # Add optional resources
        if resources:
            payload["resources"] = {
                "replicas": resources.get("replicas"),
                "resourceLabel": resources.get("resourceLabel") or resources.get("resource_label"),
                "sessionAffinity": resources.get("sessionAffinity")
                or resources.get("session_affinity"),
                "serviceWebRequestsOnRootPath": resources.get("serviceWebRequestsOnRootPath")
                or resources.get("service_web_requests_on_root_path"),
            }

        r_data = self._client.post("customApplications/", data=payload)

        # Handle async creation - the API returns 202 with Location header
        if r_data.status_code == 202:
            location = wait_for_async_resolution(self._client, r_data.headers["Location"])
            return CustomApplication.from_location(location)
        else:
            return CustomApplication.from_server_data(r_data.json())

    def get_versions(self) -> List[Dict[str, Any]]:
        """
        Get all versions of this custom application source.

        Returns
        -------
        versions : List[Dict[str, Any]]
            List of version information for this source.
        """
        r_data = self._client.get(f"{self._path}{self.id}/versions/")
        data = r_data.json().get("data", [])
        return data  # type: ignore[no-any-return]

    def get_version(self, version_id: str) -> Dict[str, Any]:
        """
        Get details of a specific version of this source.

        Parameters
        ----------
        version_id : str
            The ID of the version to retrieve.

        Returns
        -------
        version : Dict[str, Any]
            Version details including files, environment, and configuration.
        """
        r_data = self._client.get(f"{self._path}{self.id}/versions/{version_id}/")
        data = r_data.json()
        return data  # type: ignore[no-any-return]

    def update_resources(
        self,
        resource_label: Optional[str] = None,
        replicas: Optional[int] = None,
        session_affinity: Optional[bool] = None,
        service_web_requests_on_root_path: Optional[bool] = None,
    ) -> CustomApplicationSource:
        """
        Update resource configuration for this source.

        Parameters
        ----------
        resource_label : Optional[str]
            Resource bundle ID (e.g., 'cpu.small', 'cpu.large').
        replicas : Optional[int]
            Number of replicas (1-4).
        session_affinity : Optional[bool]
            Whether to enable session affinity.
        service_web_requests_on_root_path : Optional[bool]
            Whether to serve requests on root path.

        Returns
        -------
        source : CustomApplicationSource
            The updated custom application source.
        """
        resources_payload: Dict[str, Any] = {}

        if resource_label is not None:
            resources_payload["resourceLabel"] = resource_label
        if replicas is not None:
            resources_payload["replicas"] = replicas
        if session_affinity is not None:
            resources_payload["sessionAffinity"] = session_affinity
        if service_web_requests_on_root_path is not None:
            resources_payload["serviceWebRequestsOnRootPath"] = service_web_requests_on_root_path

        if not resources_payload:
            return self  # No changes to make

        payload = {"resources": resources_payload}
        r_data = self._client.patch(f"{self._path}{self.id}/", data=payload)
        return self.from_server_data(r_data.json())

    def get_file(self, version_id: str, item_id: str) -> Dict[str, Any]:
        """
        Retrieve a specific file from a version of this custom application source.

        Parameters
        ----------
        version_id : str
            The ID of the source version containing the file.
        item_id : str
            The ID of the file to download.

        Returns
        -------
        file_data : Dict[str, Any]
            The file information and content as a JSON object.

        Examples
        --------
        .. code-block:: python

            from datarobot import CustomApplicationSource

            source = CustomApplicationSource.get("source_id")

            # Get the latest version
            versions = source.get_versions()
            version_id = versions[0]["id"]

            # Get a file from that version
            version_info = source.get_version(version_id)
            item_id = version_info["items"][0]["id"]
            file_data = source.get_file(version_id, item_id)

            print(f"File name: {file_data.get('fileName', 'Unknown')}")
            if 'content' in file_data:
                print(f"Content: {file_data['content']}")
        """
        url = f"{self._path}{self.id}/versions/{version_id}/items/{item_id}/"
        r_data = self._client.get(url)
        return r_data.json()  # type: ignore[no-any-return]



#
# Copyright 2023-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from contextvars import ContextVar
from enum import Enum
from functools import wraps
import inspect
from typing import Any, Callable, Dict, Optional, TypeVar

from typing_extensions import ParamSpec

from datarobot.auth.datarobot.oauth import SyncOAuth as DataRobotOAuthClient
from datarobot.auth.identity import Identity
from datarobot.auth.session import AuthCtx

authorization_context_var: ContextVar[Dict[str, Any]] = ContextVar("authorization_context")


def set_authorization_context(authorization_context: Dict[str, Any]) -> None:
    authorization_context_var.set(authorization_context)


def get_authorization_context() -> Dict[str, Any]:
    return authorization_context_var.get()


class ToolAuth(Enum):
    OBO = "on-behalf-of"

    def __str__(self) -> str:
        return f"{self.name} ({self.value})"


P = ParamSpec("P")
T = TypeVar("T")


def is_injection_possible(func: Callable[..., Any], parameter_name: str) -> bool:
    """Check if a parameter can be injected into the function."""
    sig = inspect.signature(func)
    parameters = sig.parameters
    return parameter_name in parameters or inspect.Parameter.VAR_KEYWORD in parameters.values()


def datarobot_tool_auth(
    func: Optional[Callable[..., Any]] = None,
    *,
    type: ToolAuth = ToolAuth.OBO,
    provider: Optional[str] = None,
) -> Callable[..., Any]:
    """
    Decorator to generate and inject OAuth access token into the function.

    Parameters
    ----------
    func : Callable, optional
        The function to wrap
    type : ToolAuth, default=ToolAuth.OBO
        The type of authorization e.g., OBO (on-behalf-of).
    provider : str, optional
        The name of the OAuth provider.

    Returns
    -------
    Callable
        The decorated function with the injected token.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:

        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            if not is_injection_possible(func, "token"):
                raise ValueError(
                    f"Token injection is not possible for function {func.__name__}. "
                    f"Ensure the function has a 'token' parameter or accepts `**kwargs`."
                )
            kwargs["token"] = OAuthAccessTokenProvider().get_token(type, provider)
            return func(*args, **kwargs)

        return wrapper

    # Make decorator work also without parenthesis
    if callable(func):
        return decorator(func)

    return decorator


class OAuthAccessTokenProvider:
    """
    Manages OAuth access tokens using DataRobot-based authentication.
    """

    def __init__(self, auth_ctx: Optional[AuthCtx[Any]] = None) -> None:
        self.auth_ctx = auth_ctx or AuthCtx(**get_authorization_context())

    def _get_identity(self, provider_type: Optional[str]) -> Identity:
        """Retrieve the appropriate identity from the authentication context."""
        if not provider_type:
            if len(self.auth_ctx.identities) > 1:
                raise ValueError(
                    "The 'provider' parameter is required when there are multiple identities. "
                    "Please specify the provider to select the appropriate identity."
                )
            return self.auth_ctx.identities[0]

        for identity in self.auth_ctx.identities:
            if identity.provider_type == provider_type:
                return identity

        raise ValueError(f"No identity found for provider '{provider_type}'.")

    def get_token(self, auth_type: ToolAuth, provider_type: Optional[str] = None) -> str:
        """Get OAuth access token using the specified method."""
        identity = self._get_identity(provider_type)
        oauth_client = DataRobotOAuthClient()

        supported_auth_types = ", ".join(str(t) for t in ToolAuth)

        if auth_type == ToolAuth.OBO:
            token_data = oauth_client.refresh_access_token(
                identity_id=identity.provider_identity_id
            )
            return token_data.access_token

        raise ValueError(
            f"Unsupported tool auth type: {auth_type}. Please "
            f"use one of supported ones: {supported_auth_types}"
        )



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import os
import shutil
import tempfile
import time
from typing import Any, List, Optional, Tuple

import trafaret as t

from datarobot import errors
from datarobot._compat import Int, String, TypedDict
from datarobot.enums import DEFAULT_MAX_WAIT, EXECUTION_ENVIRONMENT_VERSION_BUILD_STATUS
from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate


class ExecutionEnvironmentVersionType(TypedDict):
    """
    Class to define ExecutionEnvironmentVersion type.
    """

    id: str
    environment_id: str
    build_status: str
    image_id: str
    label: Optional[str]
    description: Optional[str]
    created_at: Optional[str]
    docker_context_size: Optional[int]
    docker_image_size: Optional[int]
    docker_image_uri: Optional[str]


class ExecutionEnvironmentVersion(APIObject):
    """A version of a DataRobot execution environment.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        the id of the execution environment version
    environment_id: str
        the id of the execution environment the version belongs to
    build_status: str
        the status of the execution environment version build
    image_id: str
        The Docker image ID of the environment version.
    label: Optional[str]
        the label of the execution environment version
    description: Optional[str]
        the description of the execution environment version
    created_at: Optional[str]
        ISO-8601 formatted timestamp of when the execution environment version was created
    docker_context_size: Optional[int]
        The size of the uploaded Docker context in bytes if available or None if not
    docker_image_size: Optional[int]
        The size of the built Docker image in bytes if available or None if not
    docker_image_uri: Optional[str]
        The URI that the source Docker image execution environment version is based on.
        Set to None if there is not one provided.
    """

    _path = "executionEnvironments/{}/versions/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("environment_id"): String(),
            t.Key("build_status"): String(),
            t.Key("image_id"): String(),
            t.Key("label", optional=True): t.Or(String(max_length=50, allow_blank=True), t.Null()),
            t.Key("description", optional=True): t.Or(
                String(max_length=10000, allow_blank=True), t.Null()
            ),
            t.Key("created", optional=True) >> "created_at": String(),
            t.Key("docker_context_size", optional=True): t.Or(Int(), t.Null()),
            t.Key("docker_image_size", optional=True): t.Or(Int(), t.Null()),
            t.Key("source_docker_image_uri", optional=True)
            >> "docker_image_uri": t.Or(String(allow_blank=True), t.Null()),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs: Any):
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.label or self.id!r})"

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        id: str,
        environment_id: str,
        build_status: str,
        image_id: str,
        label: Optional[str] = None,
        description: Optional[str] = None,
        created_at: Optional[str] = None,
        docker_context_size: Optional[int] = None,
        docker_image_size: Optional[int] = None,
        docker_image_uri: Optional[str] = None,
    ) -> None:
        self.id = id
        self.environment_id = environment_id
        self.build_status = build_status
        self.image_id = image_id
        self.label = label
        self.description = description
        self.created_at = created_at
        self.docker_context_size = docker_context_size
        self.docker_image_size = docker_image_size
        self.docker_image_uri = docker_image_uri

    @classmethod
    def create(
        cls,
        execution_environment_id: str,
        docker_context_path: Optional[str] = None,
        docker_image_uri: Optional[str] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        max_wait: Optional[int] = DEFAULT_MAX_WAIT,
    ) -> "ExecutionEnvironmentVersion":
        """Create an execution environment version.

        .. versionadded:: v2.21

        Parameters
        ----------
        execution_environment_id: str
            the id of the execution environment
        docker_context_path: Optional[str]
            The path to a Docker context archive or folder. This parameter has lower priority
            than `docker_image_uri` if they are both provided.
        docker_image_uri: Optional[str]
            The `docker_image_uri` to be used as an environment.
            It has priority over the `docker_context_path`. If both are provided,
            the environment is created from `docker_image_uri`, and context is uploaded for
            information purposes.
        label: Optional[str]
            A human-readable string to label the version.
        description: Optional[str]
            execution environment version description
        max_wait: Optional[int]
            max time to wait for a final build status ("success" or "failed").
            If set to None - method will return without waiting.

        Returns
        -------
        ExecutionEnvironmentVersion
            created execution environment version

        Raises
        ------
        datarobot.errors.AsyncTimeoutError
            if version did not reach final state during timeout seconds
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        if not docker_context_path and not docker_image_uri:
            raise ValueError("Either docker_context_path or docker_image_url must be provided.")

        url = cls._path.format(execution_environment_id)
        payload = {"label": label, "description": description}
        if docker_image_uri:
            payload["dockerImageUri"] = docker_image_uri

        should_cleanup = False

        if docker_context_path:
            if os.path.isdir(docker_context_path):
                with tempfile.NamedTemporaryFile(
                    prefix="docker_context_", suffix=".zip"
                ) as temp_zip_file:
                    temp_zip_file_path = temp_zip_file.name
                archive_base_name = os.path.splitext(temp_zip_file_path)[0]
                archive_path = shutil.make_archive(archive_base_name, "zip", docker_context_path)
                should_cleanup = True
            else:
                archive_path = docker_context_path

            try:
                with open(archive_path, "rb") as docker_context_file:
                    response = cls._client.build_request_with_file(
                        form_data=payload,
                        fname=os.path.basename(archive_path),
                        file_field_name="docker_context",
                        filelike=docker_context_file,
                        url=url,
                        method="post",
                    )
            finally:
                if should_cleanup and archive_path and os.path.exists(archive_path):
                    os.remove(archive_path)
        else:
            response = cls._client.post(url, data=payload)

        version_id = response.json()["id"]

        if max_wait is None:
            return cls.get(execution_environment_id, version_id)
        return cls._await_final_build_status(execution_environment_id, version_id, max_wait)

    @classmethod
    def list(
        cls, execution_environment_id: str, build_status: Optional[str] = None
    ) -> List["ExecutionEnvironmentVersion"]:
        """List execution environment versions available to the user.
        .. versionadded:: v2.21

        Parameters
        ----------
        execution_environment_id: str
            the id of the execution environment
        build_status: Optional[str]
            build status of the execution environment version to filter by.
            See datarobot.enums.EXECUTION_ENVIRONMENT_VERSION_BUILD_STATUS for valid options

        Returns
        -------
        List[ExecutionEnvironmentVersion]
            a list of execution environment versions.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = cls._path.format(execution_environment_id)
        data = unpaginate(url, {"build_status": build_status}, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, execution_environment_id: str, version_id: str) -> "ExecutionEnvironmentVersion":
        """Get execution environment version by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        execution_environment_id: str
            the id of the execution environment
        version_id: str
            the id of the execution environment version to retrieve

        Returns
        -------
        ExecutionEnvironmentVersion
            retrieved execution environment version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = cls._path.format(execution_environment_id)
        path = f"{url}{version_id}/"
        return cls.from_location(path)

    def download(self, file_path: str) -> None:
        """Download execution environment version.

        .. versionadded:: v2.21

        Parameters
        ----------
        file_path: str
            path to create a file with execution environment version content

        Returns
        -------
        ExecutionEnvironmentVersion
            retrieved execution environment version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = self._path.format(self.environment_id)
        path = f"{url}{self.id}/download/"

        response = self._client.get(path)
        with open(file_path, "wb") as f:
            f.write(response.content)

    def get_build_log(self) -> Tuple[str, str]:
        """Get execution environment version build log and error.

        .. versionadded:: v2.21

        Returns
        -------
        Tuple[str, str]
            retrieved execution environment version build log and error.
            If there is no build error - None is returned.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = self._path.format(self.environment_id)
        path = f"{url}{self.id}/buildLog/"
        result = self._client.get(path).json()
        log = result["log"]
        error = result["error"]
        if error == "":
            error = None
        return log, error

    def refresh(self) -> None:
        """Update execution environment version with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        base_url = self._path.format(self.environment_id)
        url = f"{base_url}{self.id}/"
        response = self._client.get(url)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    @classmethod
    def _await_final_build_status(
        cls, execution_environment_id: str, version_id: str, max_wait: int
    ) -> "ExecutionEnvironmentVersion":
        """Awaits until an execution environment version gets to a final state.

        Parameters
        ----------
        execution_environment_id: str
            the id of the execution environment
        version_id: str
            the id of the execution environment version to retrieve
        max_wait: int
            max time to wait in seconds

        Returns
        -------
        ExecutionEnvironmentVersion
            execution environment version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        datarobot.errors.AsyncTimeoutError
            if version did not reach final state during timeout seconds
        """
        timeout_at = time.time() + max_wait
        while True:
            version = cls.get(execution_environment_id, version_id)
            if version.build_status in EXECUTION_ENVIRONMENT_VERSION_BUILD_STATUS.FINAL_STATUSES:
                break
            if time.time() >= timeout_at:
                raise errors.AsyncTimeoutError(
                    "Timeout while waiting for environment version to be built. Timeout: {}, "
                    "current state: {}, environment id: {}, environment version id: {}".format(
                        max_wait,
                        version.build_status,
                        version.id,
                        version.environment_id,
                    )
                )
            time.sleep(5)
        return version



#
# Copyright 2024-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t
from typing_extensions import TypedDict

from datarobot.enums import ToolCallAccuracyArgumentComparisonMode, enum_to_list
from datarobot.models.api_object import APIObject
from datarobot.models.genai.playground_moderation_configuration import (
    ModerationConfigurationWithoutId,
    moderation_configuration_without_id,
)
from datarobot.utils import from_api


def _get_obj(value: Union[Dict[str, Any], None], obj: Any) -> Any:
    return obj.from_server_data(value) if value is not None else None


class ToolCallAccuracySettingsDict(TypedDict):
    """Dictionary representation of the tool call accuracy settings."""

    argument_comparison: ToolCallAccuracyArgumentComparisonMode


class ExtraMetricSettingsDict(TypedDict):
    """Dictionary representation of the extra metrics settings."""

    tool_call_accuracy: Optional[ToolCallAccuracySettingsDict]


class ToolCallAccuracySettings(APIObject):
    """An object that represents the settings for tool call accuracy metrics.

    Attributes
    ----------
    argument_comparison : ToolCallAccuracyArgumentComparisonMode
        The mode for comparing arguments in tool calls.
    """

    _converter = t.Dict(
        {
            t.Key("argument_comparison"): t.Or(
                t.Enum(*enum_to_list(ToolCallAccuracyArgumentComparisonMode))
            ),
        }
    ).ignore_extra("*")

    def __init__(self, argument_comparison: ToolCallAccuracyArgumentComparisonMode):
        self.argument_comparison = argument_comparison

    def to_dict(self) -> ToolCallAccuracySettingsDict:
        return ToolCallAccuracySettingsDict(
            argument_comparison=self.argument_comparison,
        )


class ExtraMetricSettings(APIObject):
    """An object that represents additional settings for OOTB metrics.

    Attributes
    ----------
    tool_call_accuracy : Optional[ToolCallAccuracySettings]
        The settings for tool call accuracy metrics, if applicable.
    """

    _converter = t.Dict(
        {
            t.Key("tool_call_accuracy"): t.Or(ToolCallAccuracySettings._converter, t.Null),
        }
    ).ignore_extra("*")

    def __init__(self, tool_call_accuracy: Optional[ToolCallAccuracySettings]):
        self.tool_call_accuracy = tool_call_accuracy

    @classmethod
    def from_data(cls, data: Union[Dict[str, Any], List[Dict[str, Any]]]) -> ExtraMetricSettings:
        """Properly convert composition classes."""
        converted_data = cls._converter.check(from_api(data))
        tool_call_accuracy = converted_data.get("tool_call_accuracy")
        converted_data["tool_call_accuracy"] = (
            ToolCallAccuracySettings.from_data(tool_call_accuracy) if tool_call_accuracy else None
        )
        return cls(**converted_data)

    def to_dict(self) -> ExtraMetricSettingsDict:
        result = ExtraMetricSettingsDict(
            tool_call_accuracy=(
                self.tool_call_accuracy.to_dict() if self.tool_call_accuracy is not None else None
            ),
        )
        return result


class OOTBMetricConfigurationRequest(APIObject):
    """
    An object that represents a request for an out-of-the-box (OOTB) metric.

    Attributes
    ----------
    ootb_metric_name : str
        The DataRobot-defined name of the OOTB metric.
    custom_ootb_metric_name : Optional[str]
        The custom OOTB metric name chosen by the user.
    llm_id : Optional[str]
        The ID of the LLM to use for `correctness` and `faithfulness` metrics.
    custom_model_llm_validation_id : Optional[str]
        The ID of the custom model LLM validation (if using a custom model LLM).
    moderation_configuration : Optional[ModerationConfigurationWithoutId]
        The moderation configuration to be associated with the OOTB metric.
    extra_metric_settings : Optional[ExtraMetricSettings]
        Additional settings for the OOTB metric.
    """

    _converter = t.Dict(
        {
            t.Key("ootb_metric_name"): t.String,
            t.Key("custom_ootb_metric_name", optional=True): t.Or(t.String, t.Null),
            t.Key("llm_id", optional=True): t.Or(t.String, t.Null),
            t.Key("custom_model_llm_validation_id", optional=True): t.Or(t.String, t.Null),
            t.Key("moderation_configuration", optional=True): t.Or(
                moderation_configuration_without_id, t.Null
            ),
            t.Key("extra_metric_settings", optional=True): t.Or(
                ExtraMetricSettings._converter, t.Null
            ),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        ootb_metric_name: str,
        custom_ootb_metric_name: Optional[str] = None,
        llm_id: Optional[str] = None,
        custom_model_llm_validation_id: Optional[str] = None,
        moderation_configuration: Optional[Dict[str, Any]] = None,
        extra_metric_settings: Optional[ExtraMetricSettings] = None,
    ):

        self.ootb_metric_name = ootb_metric_name
        self.custom_ootb_metric_name = custom_ootb_metric_name
        self.llm_id = llm_id
        self.custom_model_llm_validation_id = custom_model_llm_validation_id
        self.moderation_configuration = _get_obj(
            moderation_configuration, ModerationConfigurationWithoutId
        )
        self.extra_metric_settings = extra_metric_settings

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(ootb_metric_name={self.ootb_metric_name})"

    @classmethod
    def from_data(
        cls, data: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> OOTBMetricConfigurationRequest:
        """Properly convert composition classes."""
        converted_data = cls._converter.check(from_api(data))
        extra_metric_settings = converted_data.get("extra_metric_settings")
        converted_data["extra_metric_settings"] = (
            ExtraMetricSettings.from_data(extra_metric_settings) if extra_metric_settings else None
        )
        return cls(**converted_data)

    def to_dict(self, uppercase_llm_key: bool = False) -> Dict[str, Any]:
        custom_model_llm_validation_id_key = (
            "custom_modelLLM_validation_id"
            if uppercase_llm_key
            else "custom_model_llm_validation_id"
        )
        return {
            "ootb_metric_name": self.ootb_metric_name,
            "custom_ootb_metric_name": self.custom_ootb_metric_name,
            "llm_id": self.llm_id,
            custom_model_llm_validation_id_key: self.custom_model_llm_validation_id,
            "moderation_configuration": (
                self.moderation_configuration.to_dict()
                if self.moderation_configuration is not None
                else None
            ),
            "extra_metric_settings": (
                self.extra_metric_settings.to_dict()
                if self.extra_metric_settings is not None
                else None
            ),
        }


class OOTBMetricConfigurationResponse(APIObject):
    """
    An object that represents a single OOTB metric.

    Attributes
    ----------
    ootb_metric_configuration_id : str
        The OOTB metric configuration ID.
    ootb_metric_name : str
        The DataRobot-defined name of the OOTB metric.
    execution_status : str
        The execution status of the OOTB metric.
    custom_ootb_metric_name : Optional[str]
        The custom OOTB metric name chosen by the user.
    llm_id : Optional[str]
        The ID of the LLM to use for `correctness` and `faithfulness` metrics.
    custom_model_llm_validation_id : Optional[str]
        The ID of the custom model LLM validation (if using a custom model LLM).
    moderation_configuration : Optional[ModerationConfigurationWithoutId]
        The moderation configuration to be associated with the OOTB metric.
    error_message : Optional[str]
        The error message associated with the OOTB metric configuration.
    error_resolution: Optional[list[str]]
        The error type associated with the insight error status.
    extra_metric_settings: Optional[Dict[str, Any]]
        Additional settings for the OOTB metric.
    """

    _converter = t.Dict(
        {
            t.Key("ootb_metric_configuration_id"): t.String,
            t.Key("ootb_metric_name"): t.String,
            t.Key("execution_status"): t.String,
            t.Key("custom_ootb_metric_name", optional=True): t.Or(t.String, t.Null),
            t.Key("llm_id", optional=True): t.Or(t.String, t.Null),
            t.Key("custom_model_llm_validation_id", optional=True): t.Or(t.String, t.Null),
            t.Key("moderation_configuration", optional=True): t.Or(
                moderation_configuration_without_id, t.Null
            ),
            t.Key("error_message", optional=True): t.Or(t.String(), t.Null),
            t.Key("error_resolution", optional=True): t.Or(t.List(t.String()), t.Null),
            t.Key("extra_metric_settings", optional=True): t.Or(
                ExtraMetricSettings._converter, t.Null
            ),
        }
    ).ignore_extra("*")
    _path = "api/v2/genai/ootbMetricConfigurations"

    def __init__(
        self,
        ootb_metric_configuration_id: str,
        ootb_metric_name: str,
        execution_status: str,
        custom_ootb_metric_name: Optional[str] = None,
        llm_id: Optional[str] = None,
        custom_model_llm_validation_id: Optional[str] = None,
        moderation_configuration: Optional[Dict[str, Any]] = None,
        error_message: Optional[str] = None,
        error_resolution: Optional[List[str]] = None,
        extra_metric_settings: Optional[ExtraMetricSettings] = None,
    ):
        self.ootb_metric_configuration_id = ootb_metric_configuration_id
        self.ootb_metric_name = ootb_metric_name
        self.execution_status = execution_status
        self.custom_ootb_metric_name = custom_ootb_metric_name
        self.llm_id = llm_id
        self.custom_model_llm_validation_id = custom_model_llm_validation_id
        self.moderation_configuration = _get_obj(
            moderation_configuration, ModerationConfigurationWithoutId
        )
        self.error_message = error_message
        self.error_resolution = error_resolution
        self.extra_metric_settings = extra_metric_settings

    @classmethod
    def from_data(
        cls, data: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> OOTBMetricConfigurationResponse:
        """Properly convert composition classes."""
        converted_data = cls._converter.check(from_api(data))
        extra_metric_settings = converted_data.get("extra_metric_settings")
        converted_data["extra_metric_settings"] = (
            ExtraMetricSettings.from_data(extra_metric_settings) if extra_metric_settings else None
        )
        return cls(**converted_data)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(ootb_metric_name={self.ootb_metric_name})"

    @classmethod
    def get(cls, ootb_metric_configuration_id: str) -> OOTBMetricConfigurationResponse:
        """Get the OOTB metric configuration."""
        response_data = cls._client.get(
            f"{cls._client.domain}/{cls._path}/{ootb_metric_configuration_id}/"
        )
        return cls.from_server_data(response_data.json())

    def delete(self) -> None:
        """Delete the OOTB metric configuration."""
        url = f"{self._client.domain}/{self._path}/{self.ootb_metric_configuration_id}/"
        self._client.delete(url)


class PlaygroundOOTBMetricConfiguration(APIObject):
    """OOTB metric configurations for a playground.

    Attributes
    ----------
    ootb_metric_configurations: (List[OOTBMetricConfigurationResponse]): The list of the OOTB metric configurations.
    """

    _converter = t.Dict(
        {
            t.Key("ootb_metric_configurations"): t.List(OOTBMetricConfigurationResponse._converter),
        }
    ).ignore_extra("*")
    path = "api/v2/genai/playgrounds/{playground_id}/ootbMetricConfigurations"

    def __init__(
        self,
        ootb_metric_configurations: List[Dict[str, Any]],
    ):
        self.ootb_metric_configurations = [
            OOTBMetricConfigurationResponse.from_server_data(config)
            for config in ootb_metric_configurations
        ]

    @classmethod
    def get(cls, playground_id: str) -> PlaygroundOOTBMetricConfiguration:
        """Get OOTB metric configurations for the playground."""
        response_data = cls._client.get(
            f"{cls._client.domain}/{cls.path.format(playground_id=playground_id)}/"
        )
        return cls.from_server_data(response_data.json())

    @classmethod
    def create(
        cls, playground_id: str, ootb_metric_configurations: list[OOTBMetricConfigurationRequest]
    ) -> PlaygroundOOTBMetricConfiguration:
        """Create a new OOTB metric configurations."""
        payload = {
            "ootb_metric_configurations": [
                config.to_dict(uppercase_llm_key=True) for config in ootb_metric_configurations
            ],
        }
        url = f"{cls._client.domain}/{cls.path.format(playground_id=playground_id)}/"
        response_data = cls._client.post(url, data=payload)
        return cls.from_server_data(response_data.json())



#
# Copyright 2024-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import trafaret as t
from trafaret.contrib.rfc_3339 import DateTime

from datarobot.enums import (
    DEFAULT_MAX_WAIT,
    ChunkServiceDialect,
    ChunkingPartitionMethod,
    ChunkingStrategy,
    enum_to_list,
)
from datarobot.models.api_object import APIObject
from datarobot.utils import from_api
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


class DatasetProps(APIObject):
    """
    The dataset props for a catalog dataset.

    Attributes
    ----------
    dataset_id : str
       The ID of the AI Catalog dataset.
    dataset_version_id : str
       The ID of the AI Catalog dataset version.

    """

    _converter = t.Dict(
        {
            t.Key("dataset_id"): t.String,
            t.Key("dataset_version_id"): t.String,
        }
    ).allow_extra("*")

    def __init__(
        self,
        dataset_id: str,
        dataset_version_id: str,
    ):
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id


class DatasetInfo(APIObject):
    """
    The dataset information.

    Attributes
    ----------
    total_rows : str
         The total number of rows in the dataset.
    source_size : str
         The size of the dataset.
    estimated_size_per_row : str
         The estimated size per row.
    columns : str
         The list of column names in the dataset.
    dialect : str
         The sql dialect associated with the dataset (e.g., Snowflake, BigQuery, Spark).
    version : int
         The version of the dataset definition information.
    data_store_id : str
         The ID of the data store.
    data_source_id : str
         The ID of the data request used to generate sampling and metadata.
    """

    _converter = t.Dict(
        {
            t.Key("total_rows"): t.Int,
            t.Key("source_size"): t.Int,
            t.Key("estimated_size_per_row"): t.Int,
            t.Key("columns"): t.List(t.String),
            t.Key("dialect"): t.Enum(*enum_to_list(ChunkServiceDialect)),
            t.Key("version"): t.Int,
            t.Key("data_store_id", optional=True): t.Or(t.String, t.Null),
            t.Key("data_source_id", optional=True): t.Or(t.String, t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        total_rows: int,
        source_size: int,
        estimated_size_per_row: int,
        columns: List[str],
        dialect: ChunkServiceDialect,
        version: int,
        data_store_id: Optional[str] = None,
        data_source_id: Optional[str] = None,
    ):
        self.total_rows = total_rows
        self.source_size = source_size
        self.estimated_size_per_row = estimated_size_per_row
        self.columns = columns
        self.dialect = dialect
        self.version = version
        self.data_store_id = data_store_id
        self.data_source_id = data_source_id


class DynamicDatasetProps(APIObject):
    """
    The dataset props for a dynamic dataset.

    Attributes
    ----------
    credentials_id : str
        The ID of the credentials.
    """

    _converter = t.Dict(
        {
            t.Key("credentials_id"): t.String,
        }
    ).allow_extra("*")

    def __init__(
        self,
        credentials_id: str,
    ):
        self.credentials_id = credentials_id


class DatasetDefinitionInfoHistory(APIObject):
    """
    Dataset definition info history for the dataset definition.

    Attributes
    ----------
    id : str
        The ID of the dataset definition information history.
    dataset_info : DatasetInfo
        The versioned information about the dataset.
    """

    _path = "datasetDefinitions/{}/versions/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("dataset_info"): DatasetInfo._converter,
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        dataset_info: DatasetInfo,
    ):
        self.id = id
        self.dataset_info = dataset_info

    @classmethod
    def from_data(
        cls, data: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> DatasetDefinitionInfoHistory:
        """Properly convert composition classes."""
        converted_data = cls._converter.check(from_api(data))
        converted_data["dataset_info"] = DatasetInfo(**converted_data["dataset_info"])

        return cls(**converted_data)

    @classmethod
    def list(cls, dataset_definition_id: str) -> List[DatasetDefinitionInfoHistory]:
        """
        List all dataset definition info history

        Parameters
        ----------
        dataset_definition_id: str
            The ID of the dataset definition.

        Returns
        -------
        A list of DatasetDefinitionInfoHistory
        """

        path = cls._path.format(dataset_definition_id)
        data = unpaginate(path, None, cls._client)
        return [cls.from_server_data(item) for item in data]


class DatasetDefinition(APIObject):
    """
    Dataset definition that holds information of dataset for API responses.

    Attributes
    ----------
    id : str
        The ID of the data source definition.
    creator_user_id : str
      The ID of the user.
    dataset_props : DatasetProps
      The properties of the dataset in catalog.
    dynamic_dataset_props : DynamicDatasetProps
      The properties of the dynamic dataset.
    dataset_info : DatasetInfo
      The information about the dataset.
    name: str
      The optional custom name of the dataset definition.
    """

    _path = "datasetDefinitions/"
    _path_with_id = _path + "{}/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("creator_user_id"): t.String,
            t.Key("dataset_props"): DatasetProps._converter,
            t.Key("dynamic_dataset_props", optional=True): t.Or(
                DynamicDatasetProps._converter, t.Null
            ),
            t.Key("dataset_info", optional=True): t.Or(DatasetInfo._converter, t.Null),
            t.Key("name", optional=True): t.Or(t.String, t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        creator_user_id: str,
        dataset_props: DatasetProps,
        dynamic_dataset_props: Optional[DynamicDatasetProps] = None,
        dataset_info: Optional[DatasetInfo] = None,
        name: Optional[str] = None,
    ):
        self.id = id
        self.creator_user_id = creator_user_id
        self.dataset_props = dataset_props
        self.dynamic_dataset_props = dynamic_dataset_props
        self.dataset_info = dataset_info
        self.name = name

    @classmethod
    def from_data(cls, data: Union[Dict[str, Any], List[Dict[str, Any]]]) -> DatasetDefinition:
        """Properly convert composition classes."""
        converted_data = cls._converter.check(from_api(data))
        converted_data["dataset_props"] = DatasetProps(**converted_data["dataset_props"])

        if "dataset_info" in converted_data:
            converted_data["dataset_info"] = DatasetInfo(**converted_data["dataset_info"])

        if "dynamic_dataset_props" in converted_data:
            converted_data["dynamic_dataset_props"] = DynamicDatasetProps(
                **converted_data["dynamic_dataset_props"]
            )

        return cls(**converted_data)

    @classmethod
    def create(
        cls,
        dataset_id: str,
        dataset_version_id: Optional[str] = None,
        name: Optional[str] = None,
        credentials_id: Optional[str] = None,
    ) -> DatasetDefinition:
        """
        Create a dataset definition.

        In order to create a dataset definition, you must first have an existing dataset in the Data Registry.
        A dataset can be uploaded using ``dr.Dataset.create_from_file`` if you have a file for example

        If you have an existing dataset in the Data Registry:

            - Retrieve the dataset ID by the canonical name via:

                - ``[cr for cr in dr.Dataset.list() if cr.name == <name>][0].id``
            - Retrieve the dataset version ID by the name via:

                - ``[cr for cr in dr.Dataset.list() if cr.name == <name>][0].version_id``

        Parameters
        ----------
        dataset_id : str
            The ID of the AI Catalog dataset.
        dataset_version_id : str
            The optional ID of the AI Catalog dataset version.
        name: str
            The optional custom name of the dataset definition.
        credentials_id: str
            The optional ID of the credentials to access the data store.


        Returns
        -------
        dataset_definition: DatasetDefinition
            An instance of a created dataset definition.

        """
        payload = {
            "dataset_id": dataset_id,
        }

        if dataset_version_id is not None:
            payload["dataset_version_id"] = dataset_version_id

        if name is not None:
            payload["name"] = name

        if credentials_id is not None:
            payload["credentials_id"] = credentials_id

        response = cls._client.post(cls._path, data=payload)
        data = response.json()
        return cls.from_server_data(data)

    @classmethod
    def get(cls, dataset_definition_id: str, version: Optional[int] = None) -> DatasetDefinition:
        """
        Retrieve a specific dataset definition metadata.

        Parameters
        ----------
        dataset_definition_id: str
            The ID of the dataset definition.
        version: Optional[int]
            The version of the dataset definition information. If not provided, the latest version will be used.

        Returns
        -------
        dataset_definition_id : DatasetDefinition
            The queried instance.
        """
        path = cls._path_with_id.format(dataset_definition_id)
        query_params = {}
        if version is not None:
            query_params["version"] = version
        response = cls._client.get(path, params=query_params)
        return cls.from_server_data(response.json())

    @classmethod
    def delete(cls, dataset_definition_id: str) -> None:
        """
        Delete a specific dataset definition

        Parameters
        ----------
        dataset_definition_id: str
            The ID of the dataset definition.

        """
        path = cls._path_with_id.format(dataset_definition_id)
        cls._client.delete(path)

    @classmethod
    def list(cls) -> List[DatasetDefinition]:
        """
        List all dataset definitions

        Returns
        -------
        A list of DatasetDefinition
        """
        path = cls._path
        data = unpaginate(path, None, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def analyze(cls, dataset_definition_id: str, max_wait: int = DEFAULT_MAX_WAIT) -> None:
        """
        Analyze a specific dataset definition

        Parameters
        ----------
        dataset_definition_id: str
            The ID of the dataset definition.
        max_wait: Optional[int]
            Time in seconds after which analyze is considered unsuccessful

        """
        path = cls._path_with_id.format(dataset_definition_id) + "analyze/"
        resp = cls._client.post(path, data={})
        if resp.status_code == 204:
            return
        elif resp.status_code == 202:
            # ignore new location since we return None
            wait_for_async_resolution(cls._client, resp.headers["Location"], max_wait)

    @classmethod
    def list_versions(cls, dataset_definition_id: str) -> List[DatasetDefinitionInfoHistory]:
        """
        List all info history of the dataset definition.

        Parameters
        ----------
        dataset_definition_id: str
            The ID of the dataset definition.

        Returns
        -------
        A list of DatasetDefinitionInfoHistory
        """
        return DatasetDefinitionInfoHistory.list(dataset_definition_id)


class RowsChunkDefinition(APIObject):
    """
    The rows chunk information.

    Attributes
    ----------
    order_by_columns : List[str]
         List of the sorting column names.
    is_descending_order : bool
         The sorting order. Defaults to False, ordering from smallest to largest.
    target_column : str
        The target column.
    target_class : str
        For binary target, one of the possible values. For zero inflated, will be '0'.
    user_group_column : str
        The user group column.
    datetime_partition_column : str
        The datetime partition column name used in OTV projects.
    otv_validation_start_date : datetime.datetime
        The start date for the validation set.
    otv_validation_end_date : datetime.datetime
        The end date for the validation set.
    otv_training_end_date : datetime.datetime
        The end date for the training set.
    otv_latest_timestamp : datetime.datetime
        The latest timestamp, this field is auto generated.
    otv_earliest_timestamp : datetime.datetime
        The earliest timestamp, this field is auto generated.
    otv_validation_downsampling_pct : float
        The percentage of the validation set to downsample, this field is auto generated.
    """

    _converter = t.Dict(
        {
            t.Key("order_by_columns"): t.List(t.String, min_length=0),
            t.Key("is_descending_order"): t.Bool,
            t.Key("target_column", optional=True): t.Or(t.String, t.Null),
            t.Key("target_class", optional=True): t.Or(t.String, t.Null),
            t.Key("user_group_column", optional=True): t.Or(t.String, t.Null),
            t.Key("datetime_partition_column", optional=True): t.Or(t.String, t.Null),
            t.Key("otv_validation_start_date", optional=True): t.Or(DateTime(), t.Null),
            t.Key("otv_validation_end_date", optional=True): t.Or(DateTime(), t.Null),
            t.Key("otv_training_end_date", optional=True): t.Or(DateTime(), t.Null),
            t.Key("otv_latest_timestamp", optional=True): t.Or(DateTime(), t.Null),
            t.Key("otv_earliest_timestamp", optional=True): t.Or(DateTime(), t.Null),
            t.Key("otv_validation_downsampling_pct", optional=True): t.Or(t.Float, t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        order_by_columns: List[str],
        is_descending_order: bool = False,
        target_column: Optional[str] = None,
        target_class: Optional[str] = None,
        user_group_column: Optional[str] = None,
        datetime_partition_column: Optional[str] = None,
        otv_validation_start_date: Optional[datetime] = None,
        otv_validation_end_date: Optional[datetime] = None,
        otv_training_end_date: Optional[datetime] = None,
        otv_latest_timestamp: Optional[datetime] = None,
        otv_earliest_timestamp: Optional[datetime] = None,
        otv_validation_downsampling_pct: Optional[float] = None,
    ):
        self.order_by_columns = order_by_columns
        self.is_descending_order = is_descending_order
        self.target_column = target_column
        self.target_class = target_class
        self.user_group_column = user_group_column
        self.datetime_partition_column = datetime_partition_column
        self.otv_validation_start_date = otv_validation_start_date
        self.otv_validation_end_date = otv_validation_end_date
        self.otv_training_end_date = otv_training_end_date
        self.otv_latest_timestamp = otv_latest_timestamp
        self.otv_earliest_timestamp = otv_earliest_timestamp
        self.otv_validation_downsampling_pct = otv_validation_downsampling_pct


class FeaturesChunkDefinition(APIObject):
    """The features chunk information."""

    _converter = t.Dict({}).allow_extra("*")

    def __init__(self) -> None:
        pass


class ChunkDefinitionStats(APIObject):
    """
    The chunk stats information.

    Attributes
    ----------
    expected_chunk_size: int
        The expected chunk size, this field is auto generated.
    number_of_rows_per_chunk: int
        The number of rows per chunk, this field is auto generated.
    total_number_of_chunks: int
        The total number of chunks, this field is auto generated.
    """

    _converter = t.Dict(
        {
            t.Key("expected_chunk_size"): t.Int,
            t.Key("number_of_rows_per_chunk"): t.Int,
            t.Key("total_number_of_chunks"): t.Int,
        }
    ).allow_extra("*")

    def __init__(
        self,
        expected_chunk_size: int,
        number_of_rows_per_chunk: int,
        total_number_of_chunks: int,
    ):
        self.expected_chunk_size = expected_chunk_size
        self.number_of_rows_per_chunk = number_of_rows_per_chunk
        self.total_number_of_chunks = total_number_of_chunks


class ChunkDefinition(APIObject):
    """
    The chunk information.

    Attributes
    ----------
    id : str
        The ID of the chunk entity.
    dataset_definition_id : str
        The ID of the dataset definition.
    dataset_definition_info_version: int
        The version of the dataset definition information.
    name : str
        The name of the chunk entity.
    is_readonly : bool
        The read only flag.
    partition_method : str
        The partition method used to create chunks, either 'random', 'stratified', or 'date'.
    chunking_strategy_type : str
        The chunking strategy type, either 'features' or 'rows'.
    chunk_definition_stats : ChunkDefinitionStats
        The chunk stats information.
    rows_chunk_definition : RowsChunkDefinition
        The rows chunk information.
    features_chunk_definition : FeaturesChunkDefinition
        The features chunk information.
    """

    _path = "datasetDefinitions/{}/chunkDefinitions/"
    _path_with_id = _path + "{}/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("dataset_definition_id"): t.String,
            t.Key("dataset_definition_info_version"): t.Int,
            t.Key("name"): t.String,
            t.Key("is_readonly"): t.Bool,
            t.Key("partition_method"): t.Enum(*enum_to_list(ChunkingPartitionMethod)),
            t.Key("chunking_strategy_type"): t.Enum(*enum_to_list(ChunkingStrategy)),
            t.Key("chunk_definition_stats", optional=True): ChunkDefinitionStats._converter,
            t.Key("rows_chunk_definition", optional=True): RowsChunkDefinition._converter,
            t.Key("features_chunk_definition", optional=True): FeaturesChunkDefinition._converter,
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        dataset_definition_id: str,
        dataset_definition_info_version: int,
        name: str,
        is_readonly: bool,
        partition_method: ChunkingPartitionMethod,
        chunking_strategy_type: ChunkingStrategy,
        chunk_definition_stats: Optional[ChunkDefinitionStats] = None,
        rows_chunk_definition: Optional[RowsChunkDefinition] = None,
        features_chunk_definition: Optional[FeaturesChunkDefinition] = None,
    ):
        self.id = id
        self.dataset_definition_id = dataset_definition_id
        self.dataset_definition_info_version = dataset_definition_info_version
        self.name = name
        self.is_readonly = is_readonly
        self.partition_method = partition_method
        self.chunking_strategy_type = chunking_strategy_type
        self.chunk_definition_stats = chunk_definition_stats
        self.rows_chunk_definition = rows_chunk_definition
        self.features_chunk_definition = features_chunk_definition

    @classmethod
    def from_data(cls, data: Union[Dict[str, Any], List[Dict[str, Any]]]) -> ChunkDefinition:
        """Properly convert composition classes."""
        converted_data = cls._converter.check(from_api(data))

        if "chunk_definition_stats" in converted_data:
            converted_data["chunk_definition_stats"] = ChunkDefinitionStats(
                **converted_data["chunk_definition_stats"]
            )

        if "rows_chunk_definition" in converted_data:
            converted_data["rows_chunk_definition"] = RowsChunkDefinition(
                **converted_data["rows_chunk_definition"]
            )

        if "features_chunk_definition" in converted_data:
            converted_data["features_chunk_definition"] = FeaturesChunkDefinition(
                **converted_data["features_chunk_definition"]
            )

        return cls(**converted_data)

    @classmethod
    def create(
        cls,
        dataset_definition_id: str,
        name: Optional[str] = None,
        partition_method: Optional[ChunkingPartitionMethod] = ChunkingPartitionMethod.RANDOM,
        chunking_strategy_type: Optional[ChunkingStrategy] = ChunkingStrategy.ROWS,
        order_by_columns: Optional[List[str]] = None,
        is_descending_order: Optional[bool] = False,
        target_column: Optional[str] = None,
        target_class: Optional[str] = None,
        user_group_column: Optional[str] = None,
        datetime_partition_column: Optional[str] = None,
        otv_validation_start_date: Optional[datetime] = None,
        otv_validation_end_date: Optional[datetime] = None,
        otv_training_end_date: Optional[datetime] = None,
    ) -> ChunkDefinition:
        """
        Create a chunk definition.

        Parameters
        ----------
        dataset_definition_id : str
            The ID of the dataset definition.
        name: str
            The optional custom name of the chunk definition.
        partition_method: str
            The partition method used to create chunks, either 'random', 'stratified', or 'date'.
        chunking_strategy_type: str
            The chunking strategy type, either 'features' or 'rows'.
        order_by_columns: List[str]
            List of the sorting column names.
        is_descending_order: bool
            The sorting order. Defaults to False, ordering from smallest to largest.
        target_column: str
            The target column.
        target_class: str
            For binary target, one of the possible values. For zero inflated, will be '0'.
        user_group_column: str
            The user group column.
        datetime_partition_column: str
            The datetime partition column name used in OTV projects.
        otv_validation_start_date: datetime.datetime
            The start date for the validation set.
        otv_validation_end_date: datetime.datetime
            The end date for the validation set.
        otv_training_end_date: datetime.datetime
            The end date for the training set.

        Returns
        -------
        chunk_definition: ChunkDefinition
            An instance of a created chunk definition.

        """
        _ = (
            name,
            order_by_columns,
            target_column,
            target_class,
            user_group_column,
            datetime_partition_column,
            otv_validation_start_date,
            otv_validation_end_date,
            otv_training_end_date,
        )

        payload = {
            "partition_method": partition_method,
            "chunking_strategy_type": chunking_strategy_type,
            "is_descending_order": is_descending_order,
        }

        optional_none_value_fields = [
            "name",
            "order_by_columns",
            "target_column",
            "target_class",
            "user_group_column",
            "datetime_partition_column",
            "otv_validation_start_date",
            "otv_validation_end_date",
            "otv_training_end_date",
        ]
        for field in optional_none_value_fields:
            value = locals().get(field)
            if value is not None:
                payload[field] = value

        path = cls._path.format(dataset_definition_id)
        response = cls._client.post(path, data=payload)
        data = response.json()
        return cls.from_server_data(data)

    @classmethod
    def get(cls, dataset_definition_id: str, chunk_definition_id: str) -> ChunkDefinition:
        """
        Retrieve a specific chunk definition metadata.

        Parameters
        ----------
        dataset_definition_id: str
            The ID of the dataset definition.
        chunk_definition_id: str
            The ID of the chunk definition.

        Returns
        -------
        chunk_definition : ChunkDefinition
            The queried instance.
        """
        path = cls._path_with_id.format(dataset_definition_id, chunk_definition_id)
        response = cls._client.get(path)
        return cls.from_server_data(response.json())

    @classmethod
    def delete(cls, dataset_definition_id: str, chunk_definition_id: str) -> None:
        """
        Delete a specific chunk definition

        Parameters
        ----------
        dataset_definition_id: str
            The ID of the dataset definition.
        chunk_definition_id: str
            The ID of the chunk definition.

        """
        path = cls._path_with_id.format(dataset_definition_id, chunk_definition_id)
        cls._client.delete(path)

    @classmethod
    def list(cls, dataset_definition_id: str) -> List[ChunkDefinition]:
        """
        List all chunk definitions

        Parameters
        ----------
        dataset_definition_id: str
            The ID of the dataset definition.

        Returns
        -------
        A list of ChunkDefinition
        """
        path = cls._path.format(dataset_definition_id)
        data = unpaginate(path, None, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def analyze(
        cls, dataset_definition_id: str, chunk_definition_id: str, max_wait: int = DEFAULT_MAX_WAIT
    ) -> None:
        """
        Analyze a specific chunk definition

        Parameters
        ----------
        dataset_definition_id: str
            The ID of the dataset definition.
        chunk_definition_id: str
            The ID of the chunk definition
        max_wait: Optional[int]
            Time in seconds after which analyze is considered unsuccessful

        """
        path = cls._path_with_id.format(dataset_definition_id, chunk_definition_id) + "analyze/"
        resp = cls._client.post(path, data={})
        if resp.status_code == 204:
            return
        elif resp.status_code == 202:
            # ignore new location since we return None
            wait_for_async_resolution(cls._client, resp.headers["Location"], max_wait)

    @classmethod
    def update(
        cls,
        chunk_definition_id: str,
        dataset_definition_id: str,
        name: Optional[str] = None,
        order_by_columns: Optional[List[str]] = None,
        is_descending_order: Optional[bool] = None,
        target_column: Optional[str] = None,
        target_class: Optional[str] = None,
        user_group_column: Optional[str] = None,
        datetime_partition_column: Optional[str] = None,
        otv_validation_start_date: Optional[datetime] = None,
        otv_validation_end_date: Optional[datetime] = None,
        otv_training_end_date: Optional[datetime] = None,
        force_update: Optional[bool] = False,
    ) -> ChunkDefinition:
        """
        Update a chunk definition.

        Parameters
        ----------
        chunk_definition_id: str
            The ID of the chunk definition.
        dataset_definition_id : str
            The ID of the dataset definition.
        name: str
            The optional custom name of the chunk definition.
        order_by_columns: List[str]
            List of the sorting column names.
        is_descending_order: bool
            The sorting order. Defaults to False, ordering from smallest to largest.
        target_column: str
            The target column.
        target_class: str
            For binary target, one of the possible values. For zero inflated, will be '0'.
        user_group_column: str
            The user group column.
        datetime_partition_column: str
            The datetime partition column name used in OTV projects.
        otv_validation_start_date: datetime.datetime
            The start date for the validation set.
        otv_validation_end_date: datetime.datetime
            The end date for the validation set.
        otv_training_end_date: datetime.datetime
            The end date for the training set.
        force_update: bool
            If True, the update will be forced in some cases. For example, update after analysis is done.

        Returns
        -------
        chunk_definition: ChunkDefinition
            An update instance of a created chunk definition.

        """
        _ = (
            name,
            order_by_columns,
            is_descending_order,
            target_column,
            target_class,
            user_group_column,
            datetime_partition_column,
            otv_validation_start_date,
            otv_validation_end_date,
            otv_training_end_date,
        )

        optional_none_value_fields = [
            "name",
            "order_by_columns",
            "is_descending_order",
            "target_column",
            "target_class",
            "user_group_column",
            "datetime_partition_column",
            "otv_validation_start_date",
            "otv_validation_end_date",
            "otv_training_end_date",
        ]
        updates = {}
        for field in optional_none_value_fields:
            value = locals().get(field)
            if value is not None:
                updates[field] = value

        payload = {
            "updates": updates,
            "operations": {
                "forceUpdate": force_update,
            },
        }

        path = cls._path_with_id.format(dataset_definition_id, chunk_definition_id)
        response = cls._client.patch(path, data=payload)
        data = response.json()
        return cls.from_server_data(data)
